/* FILE: mmdispcmds.cc                 -*-Mode: c++-*-
 *
 * Tcl commands implemented in C/C++ and supported by shell mmdispsh
 *
 */

#include <math.h>	/* fmod(), ... */
#include <stdlib.h>	/* atoi(), ... */
#include <string.h>	/* strncpy(), ... */

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "mmdispcmds.h"
#include "display.h"	/* DisplayFrame, CoordinateSystem */
#include "bitmap.h"

/* End includes */

#define MY_MESH_ARRAY_SIZE 2
static Vf_Mesh* myMeshArray[MY_MESH_ARRAY_SIZE];
static int activeMeshId = 0;
static Vf_Mesh *myMesh;
static DisplayFrame myFrame;

#define MY_BUF_SIZE 8192
static CHAR buf[MY_BUF_SIZE];

static REAL4m CoordsToAngle(CoordinateSystem coords)
{
  REAL4m rotang=0.;
  switch(coords) {
  case CS_CALCULATION_STANDARD: // Safety
  case CS_DISPLAY_STANDARD:  rotang=0.;   break;
  case CS_DISPLAY_ROT_90:    rotang=90.;  break;
  case CS_DISPLAY_ROT_180:   rotang=180.; break;
  case CS_DISPLAY_ROT_270:   rotang=270.; break;
  case CS_ILLEGAL:
  default:
    PlainError(1,"Unrecognized or illegal coordinate system (%d)"
	       " detected in CoordsToAngle() (File tkcmds.cc)\n",
	       int(coords));
    break; // Shouldn't get to here
  }
  return rotang;
}

static CoordinateSystem AngleToCoords(REAL4m angle_in_degrees)
{ // Rounds to nearest supported coordinate system rotation
  CoordinateSystem coords=CS_ILLEGAL;
  int quad=int(OC_ROUND(fmod(double(angle_in_degrees),
			     double(360.))/90.));
  if(quad<0) quad+=4;  if(quad>3) quad-=4;
  switch(quad) {
  case 0:  coords=CS_DISPLAY_STANDARD;  break;
  case 1:  coords=CS_DISPLAY_ROT_90;    break;
  case 2:  coords=CS_DISPLAY_ROT_180;   break;
  case 3:  coords=CS_DISPLAY_ROT_270;   break;
  default:
    PlainError(1,"Programming error: Illegal coordinate system (quad=%d)"
	       " occurred in AngleToCoords() (File tkcmds.cc)\n",quad);
    break; // Shouldn't get to here
  }
  return coords;
}

int ReportActiveMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s\"", argv[0]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),"%d",activeMeshId);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}


int SelectActiveMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
		"SelectActiveMesh must be called with 1 argument:"
		" Mesh Id to make active (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int meshId = atoi(argv[1]);
  if(meshId<0 || meshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		meshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int oldId = activeMeshId;
  activeMeshId = meshId;
  myMesh = myMeshArray[activeMeshId];

  myFrame.SetMesh(myMesh);

  Oc_Snprintf(buf,sizeof(buf),"%d",oldId);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int FreeMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s meshid\"", argv[0]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int meshId = atoi(argv[1]);
  if(meshId<0 || meshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		meshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  if (myMesh==myMeshArray[meshId]) myMesh=NULL; //Safety
  delete myMeshArray[meshId];
  myMeshArray[meshId]=new Vf_EmptyMesh();
  if(activeMeshId == meshId) myMesh = myMeshArray[meshId];

  return TCL_OK;
}

int CopyMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<3 || argc>6) {
    Oc_Snprintf(buf,sizeof(buf),
		"CopyMesh must be called with 2-5 arguments: "
		"original mesh id, new mesh id, "
		"subsample, flip string and clipbox"
		" (%d arguments passed)",
		argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int origmeshId = atoi(argv[1]);
  if(origmeshId<0 || origmeshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		origmeshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int newmeshId = atoi(argv[2]);
  if(newmeshId<0 || newmeshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		newmeshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  REAL8m subsample = 0.;
  if(argc>3) subsample = Nb_Atof(argv[3]);
  const char* flipstr = "x:y:z";
  if(argc>4) flipstr = argv[4];

  // Clipbox
  Nb_BoundingBox<REAL8m> clipbox;
  if(argc>5 && argv[5][0]!='\0') {
    const char* parsestr=" \t\n:";
    char* buf = new char[strlen(argv[5])+1]; // clipbox corners
    strcpy(buf,argv[5]);
    char* nexttoken=buf;
    char *token1 = Nb_StrSep(&nexttoken,parsestr);
    char *token2 = Nb_StrSep(&nexttoken,parsestr);
    char *token3 = Nb_StrSep(&nexttoken,parsestr);
    char *token4 = Nb_StrSep(&nexttoken,parsestr);
    char *token5 = Nb_StrSep(&nexttoken,parsestr);
    char *token6 = Nb_StrSep(&nexttoken,parsestr);

    myMeshArray[origmeshId]->GetPreciseRange(clipbox); // All by default
    Nb_Vec3<REAL8m> minpt,maxpt;
    clipbox.GetExtremes(minpt,maxpt);
    if(token1[0]!='\0' && strcmp("-",token1)!=0) {
      minpt.x = Nb_Atof(token1);
    }
    if(token2[0]!='\0' && strcmp("-",token2)!=0) {
      minpt.y = Nb_Atof(token2);
    }
    if(token3[0]!='\0' && strcmp("-",token3)!=0) {
      minpt.z = Nb_Atof(token3);
    }
    if(token4[0]!='\0' && strcmp("-",token4)!=0) {
      maxpt.x = Nb_Atof(token4);
    }
    if(token5[0]!='\0' && strcmp("-",token5)!=0) {
      maxpt.y = Nb_Atof(token5);
    }
    if(token6[0]!='\0' && strcmp("-",token6)!=0) {
      maxpt.z = Nb_Atof(token6);
    }
    clipbox.Set(minpt,maxpt);
  }

  // Create new mesh; if the original mesh is regular (i.e.,
  // Vf_GridVec3f or Vf_Empty), then make the new one of the
  // same type.
  Vf_Mesh* newmesh=NULL;
  const char* origtype = myMeshArray[origmeshId]->GetMeshType();
  if(strcmp(origtype,"Vf_EmptyMesh")==0) {
    newmesh = new Vf_EmptyMesh();
  } else if(strcmp(origtype,"Vf_GridVec3f")==0) {
    newmesh = new
      Vf_GridVec3f(*((Vf_GridVec3f*)myMeshArray[origmeshId]),
		   subsample,flipstr,clipbox);
  } else {
    newmesh = new
      Vf_GeneralMesh3f(*myMeshArray[origmeshId],subsample,
		       flipstr,clipbox);
  }

  // Place new mesh into myMeshArray
  if(myMeshArray[newmeshId]!=NULL) {
    if(myMesh==myMeshArray[newmeshId]) myMesh=NULL; // Safety
    delete myMeshArray[newmeshId];
  }
  myMeshArray[newmeshId] = newmesh;
  if(activeMeshId == newmeshId) myMesh = newmesh;

  return TCL_OK;
}

int DifferenceMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
		"DifferenceMesh must be called with 1 argument: "
		"Mesh Id to subtract from active mesh "
		"(%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int meshId = atoi(argv[1]);
  if(meshId<0 || meshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
	    "Invalid mesh id request: %d; should be between 0 and %d",
	    meshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Difference meshes here
  if((*myMeshArray[activeMeshId]).SubtractMesh(*myMeshArray[meshId])!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Meshes aren't compatible");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

int GetMeshType(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshType must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetMeshType();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

static int IsRegularMesh(Vf_Mesh* mesh)
{
  const char *name=mesh->GetMeshType();
  int regular=0;
  if(strcmp(name,"Vf_GridVec3f")==0 ||
     strcmp(name,"Vf_EmptyMesh")==0) {
    // Known regular meshes
    regular=1;
  } // Otherwise assume irregular
  return regular;
}

int IsRegularMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "IsRegularMesh must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int regular = IsRegularMesh(myMesh);
  Oc_Snprintf(buf,sizeof(buf),"%d",regular);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetVecColor(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetVecColor must be called with 1 argument:"
	    " A 3 element list representing a vector"
            " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int carr_size;
  CONST84 char** carr;
  if (Tcl_SplitList(interp,argv[1],&carr_size,&carr)!=TCL_OK) {
    return TCL_ERROR;
  }
  if(carr_size!=3) {
    Oc_Snprintf(buf,sizeof(buf),
		"Input list has %d != 3 elements",carr_size);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Nb_Vec3<REAL4> v(static_cast<REAL4m>(Nb_Atof(carr[0])),
		   static_cast<REAL4m>(Nb_Atof(carr[1])),
		   static_cast<REAL4m>(Nb_Atof(carr[2])));
  Tcl_Free((char *)carr);
  myFrame.GetVecColor(v);
  Oc_AutoBuf ab;
  Tcl_AppendResult(interp,ab(myFrame.GetVecColor(v)),(char *)NULL);
  return TCL_OK;
}


int GetMeshName(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshName must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetName();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshStructureInfo(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshStructureInfo must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m size=myMesh->GetSize();
  Nb_BoundingBox<REAL8> datarange;
  myMesh->GetPreciseDataRange(datarange);
  Nb_Vec3<REAL8m> minpt,maxpt;
  datarange.GetExtremes(minpt,maxpt);
  REAL8m value_min,value_max;
  const char *meshunit = myMesh->GetMeshUnit();
  myMesh->GetValueMagSpan(value_min,value_max);
  const char *valueunit = myMesh->GetValueUnit();
  if(IsRegularMesh(myMesh)) {
    if(size<1 || strcmp("Vf_GridVec3f",myMesh->GetMeshType())!=0) {
      Oc_Snprintf(buf,sizeof(buf),
		  "Regular mesh\n Mesh size: %d\n",size);
    } else {
      Vf_GridVec3f* regmesh = (Vf_GridVec3f*)myMesh;
      INT4m dimx,dimy,dimz;
      regmesh->GetDimens(dimx,dimy,dimz);
      Oc_Snprintf(buf,sizeof(buf),
		  "Regular mesh\n"
		  " Mesh size: %d\n"
		  " Dimensions: %d %d %d\n"
		  " Value magnitude span: %g to %g (in %s)\n"
		  " Mesh range: (%g,%g,%g) x (%g,%g,%g) (in %s)",
		  size,
		  dimx,dimy,dimz,
		  value_min,value_max,valueunit,
		  minpt.x,minpt.y,minpt.z,
		  maxpt.x,maxpt.y,maxpt.z,meshunit);
    }
    
  } else {
    Oc_Snprintf(buf,sizeof(buf),
		"Irregular mesh\n"
		" Size: %d\n"
		" Value magnitude span: %g to %g (in %s)\n"
		" Mesh range: (%g,%g,%g) x (%g,%g,%g) (in %s)",
		size,
		value_min,value_max,valueunit,
		minpt.x,minpt.y,minpt.z,
		maxpt.x,maxpt.y,maxpt.z,meshunit);
  }
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshTitle(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshTitle must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *title=myMesh->GetTitle();
  if(title!=NULL) strncpy(buf,title,sizeof(buf));
  else            buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int SetMeshTitle(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
            "SetMeshTitle must be called with 1 argument:"
	    " new_title (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  myMesh->SetTitle(argv[1]);
  return TCL_OK;
}

int GetMeshDescription(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshDesc must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *desc=myMesh->GetDescription();
  if(desc!=NULL) strncpy(buf,desc,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshCellSize(ClientData,Tcl_Interp *interp,
		    int argc,CONST84 char** /* argv */)
{
  // Returns a triple: xstep ystep zstep (in mesh units)
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshCellSize must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_Vec3<REAL4> dim=myMesh->GetApproximateCellDimensions();
  Oc_Snprintf(buf,sizeof(buf),"%.8g %.8g %.8g",dim.x,dim.y,dim.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshIncrement(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshIncrement must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),"%f",myMesh->GetSubsampleGrit());
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshSpatialUnitString(ClientData,Tcl_Interp *interp,
		  int argc,CONST84 char** argv)
{ // Returns string denoting mesh spatial units
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s\"", argv[0]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),"%s",myMesh->GetMeshUnit());
  if(strcmp("unknown",buf)==0) buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshRange(ClientData,Tcl_Interp *interp,
		  int argc,CONST84 char** /* argv */)
{ // Returns 6-tuple: xmin ymin zmin xmax ymax zmax, in mesh units.
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshRange must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_Vec3<REAL8> minpt,maxpt;
  Nb_BoundingBox<REAL8> range;
  myMesh->GetPreciseRange(range);
  range.GetExtremes(minpt,maxpt);
  Oc_Snprintf(buf,sizeof(buf),"%.17g %.17g %.17g %.17g %.17g %.17g",
	      minpt.x,minpt.y,minpt.z,maxpt.x,maxpt.y,maxpt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshZRange(ClientData,Tcl_Interp *interp,
		  int argc,CONST84 char** /* argv */)
{ // Returns a pair: zmin zmax.  These are in mesh units.
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshZRange must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_Vec3<REAL8> minpt,maxpt;
  Nb_BoundingBox<REAL8> range;
  myMesh->GetPreciseRange(range);
  REAL8m zmin=0.,zmax=0.;
  if(!range.IsEmpty()) {
    range.GetExtremes(minpt,maxpt);
    zmin=minpt.z;
    zmax=maxpt.z;
  }  
  Oc_Snprintf(buf,sizeof(buf),"%.17g %.17g",zmin,zmax);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetZsliceCount(ClientData,Tcl_Interp *interp,
		   int argc,CONST84 char** /* argv */)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetZsliceCount must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m slicecount = myMesh->GetZsliceCount();

  Oc_Snprintf(buf,sizeof(buf),"%d",slicecount);
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}

int GetZsliceLevels(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=3) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetZsliceLevels must be called with 2 arguments:"
	    " z-low z-high (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL8m zlow =Nb_Atof(argv[1]);
  REAL8m zhigh =Nb_Atof(argv[2]);
  INT4m islicelow,islicehigh;
  myMesh->GetZslice(zlow,zhigh,islicelow,islicehigh);

  Oc_Snprintf(buf,sizeof(buf),"%d %d",islicelow,islicehigh);
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}


int ChangeMesh(ClientData cd,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<5 || argc>6) {
    Oc_Snprintf(buf,sizeof(buf),
            "ChangeMesh must be called with 4 or 5 arguments:"
	    " mesh_filename"
	    " frame_width frame_height rotation_degrees"
	    " [zoom]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  if(activeMeshId<0 || activeMeshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
	    "PROGRAMMING ERROR: activeMeshId=%d is out-of-range: [0,%d]",
	    activeMeshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  if(myMesh!=myMeshArray[activeMeshId]) {
    Oc_Snprintf(buf,sizeof(buf),
	    "PROGRAMMING ERROR: myMesh not in sync with myMeshArray.");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  delete myMeshArray[activeMeshId];
  myMeshArray[activeMeshId]=myMesh=NULL; // Safety
  if(argv[1]==NULL || argv[1][0]=='\0') { // Null request
    myMeshArray[activeMeshId]=new Vf_EmptyMesh();
  }
  else {
    Vf_FileInput *vffreader
      = Vf_FileInput::NewReader(argv[1]);
    if(vffreader==NULL) { // Unknown file type
      myMeshArray[activeMeshId]=new Vf_EmptyMesh();
    }
    else { // Reader found for file argv[1]
      myMeshArray[activeMeshId]=vffreader->NewMesh();
      delete vffreader;
      if(myMeshArray[activeMeshId]==NULL)
	myMeshArray[activeMeshId]=new Vf_EmptyMesh(); // Safety
    }
  }
  myMesh = myMeshArray[activeMeshId];

  myFrame.SetMesh(myMesh);

  REAL4m width=static_cast<REAL4m>(Nb_Atof(argv[2]));
  REAL4m height=static_cast<REAL4m>(Nb_Atof(argv[3]));
  CoordinateSystem coords =
    AngleToCoords(static_cast<REAL4m>(Nb_Atof(argv[4])));
  REAL4m zoom=-1.;
  if(argc>5) {
    zoom=static_cast<REAL4m>(Nb_Atof(argv[5]));
  }

  // Adjust sizes to leave room for margins.
  Oc_AutoBuf ab1,ab2;
  const char *cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("misc,margin"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    REAL4m margin=static_cast<REAL4m>(Nb_Atof(cptr));
    width-=2*margin;
    height-=2*margin;
  }

  myFrame.SetCoordinates(coords);

  if(zoom>0.) myFrame.SetZoom(zoom);
  else        myFrame.SetZoom(width,height);

  UpdatePlotConfiguration(cd,interp,0,NULL);
  /// For now, just use current configuration /**/

  return TCL_OK;
}

int DrawFrame(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  /*
   * We use the following buffer to store commands for passing to
   * Tcl_Eval().  It is important that we do not use the global
   * buffer, buf, for commands evaluted by Tcl_Eval() since that is
   * used to write result strings to Tcl_AppendResult().  If the
   * contents of a buffer are overwritten while Tcl_Eval() is
   * evaluating them, bad results can happen.
   *
   * Note that similar bad things can happen if the contents of
   * cmdbuf are overwritten which Tcl_Eval() evaluates them.  This
   * means it is a bad thing for DrawFrame() to appear on the call
   * stack twice.  Take care that none of the Tcl_Eval()'s below make
   * another call to DrawFrame().
   */
  static CHAR cmdbuf[MY_BUF_SIZE];

  int error_code;
  Tcl_ResetResult(interp);
  if(argc!=3) {
    Oc_Snprintf(buf,sizeof(buf),
            "DrawFrame must be called with 2 arguments: "
	    "canvas_name SliceCompatMode (%d arguments passed)",
	    argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char* canvas=argv[1];
  int zslicecompat=atoi(argv[2]);

  Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
	      "{%s} delete all ;# update idletasks",canvas);
  /// Uncomment the "update idletasks" command to get a white
  /// screen between frames
  error_code=Tcl_Eval(interp,cmdbuf); // Clear canvas
  if(error_code!=TCL_OK) return error_code;

  // Redraw canvas
  if(zslicecompat) error_code=myFrame.Render(canvas,0);
  else error_code=myFrame.Render(canvas,1);  // Hide arrows and pixels
  if(error_code!=TCL_OK) {
    Oc_Snprintf(buf,sizeof(buf),
		"\nError from myFrame.Render(%s,0)",canvas);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return error_code;
  }

  Oc_Snprintf(cmdbuf,sizeof(cmdbuf),"InitializeSliceDisplay");
  error_code=Tcl_Eval(interp,cmdbuf);
  if(error_code!=TCL_OK) {
    Oc_Snprintf(buf,sizeof(buf),"\nError from InitializeSliceDisplay");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return error_code;
  }

  Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
	      "global DrawFrameCount; incr DrawFrameCount");
  error_code=Tcl_Eval(interp,cmdbuf); // Increment DrawFrameCount variable
  if(error_code!=TCL_OK) return error_code;

/* Adjust subsample setting to actual value.
 * Note: Any tracings on plot_config triggered as a result
 *  of the Tcl_SetVar2 calls will send the name used in the
 *  call (here, e.g., "plot_config arrow_subsample") to the
 *  trace callback, and the call will appear to be from the
 *  proc that called this routine (DrawFrame).  If that routine
 *  has not made plot_config visible through a 'global plot_config'
 *  command, then a "upvar plot_config" will *fail* in the trace
 *  callback.  To protect against this, trace callbacks on plot_config
 *  should have the global name hardcoded into the trace call, and the
 *  callback itself should use the hardcoded value instead of the
 *  "name" argument.  For example:
 *
 *            trace variable plot_config(arrow_subsample) w \
 *                "MyCallback plot_config"
 *  and
 *        proc MyCallback { globalname name elt op } {
 *                upvar #0 $globalname x
 *                ...
 *        }
 *
 *  (The "elt" value can be either hardcoded into globalname, or
 *  processed separately.)
 */
  Oc_AutoBuf ab1,ab2;
  REAL4m arrow_sample,pixel_sample;
  REAL4m arrow_sample_request,pixel_sample_request;
  myFrame.GetRequestedSubsampleRates(arrow_sample_request,
				     pixel_sample_request);
  myFrame.GetActualSubsampleRates(arrow_sample,pixel_sample);
  if(arrow_sample_request<0.) {
    arrow_sample*=-1;
    if(arrow_sample > REAL4m(-0.01)) arrow_sample = REAL4m(-0.01);
  }
  Oc_Snprintf(buf,sizeof(buf),"%g",arrow_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
	      buf,TCL_GLOBAL_ONLY);
  if(pixel_sample_request<0.) {
    pixel_sample*=-1;
    if(pixel_sample > REAL4m(-0.01)) pixel_sample = REAL4m(-0.01);
  }
  Oc_Snprintf(buf,sizeof(buf),"%g",pixel_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
	      buf,TCL_GLOBAL_ONLY);

  return TCL_OK;
}

int GetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),"GetFrameRotation must be called with"
	    " 1 argument: name of Tk's Frame Rotation variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m rotang=int(OC_ROUND(CoordsToAngle(myFrame.GetCoordinates())));

  Oc_Snprintf(buf,sizeof(buf),"%d",rotang);
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

// Following routine added by SLW May-1998, modified by MJD June-1998.
// It copies the plot configuration values from the Tcl variable
// plot_config into the corresponding Frame plot_config structures.
// Also sets quantitylist value from Frame.
int UpdatePlotConfiguration(ClientData,Tcl_Interp *interp,int,CONST84 char **)
{
  Tcl_ResetResult(interp);
  // Ignore argument list; just read plot_config() directly and update
  // all fields.
  const char *cptr; int colorcount;
  Oc_AutoBuf ab1,ab2;

  // Variables to handle "out-of-plane rotations", i.e., axis
  // transforms.
  Nb_DString flipstr("+x:+y:+z"),colorquantity;
  REAL8m phase=0.;
  BOOL invert=0;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("viewaxis"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr) && strcmp(cptr,"+x:+y:+z")!=0) {
    char axisbuf[8];
    Oc_Snprintf(axisbuf,sizeof(axisbuf),"+z,%s",cptr);
    cptr = Tcl_GetVar2(interp,ab1("view_transform"),axisbuf,
		       TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      flipstr = cptr;
    }
  }

  PlotConfiguration arrow_config,pixel_config;
  myFrame.GetPlotConfiguration(arrow_config,pixel_config);

  // Arrow plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,status"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("arrow,colorcount"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) colorcount=atoi(cptr);
  if(colorcount<0) colorcount=arrow_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("arrow,colormap"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.colormap.Setup(colorcount,cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,colorphase"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) phase=Nb_Atof(cptr);
  else                     phase=0.;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,colorinvert"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) invert=atoi(cptr);
  else                     invert=0;

  myMesh->ColorQuantityTransform(flipstr,colorquantity,phase,invert,
                                 arrow_config.colorquantity,
				 arrow_config.phase,arrow_config.invert);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,autosample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) arrow_config.autosample=atoi(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.subsample=static_cast<REAL4m>(Nb_Atof(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,size"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.size=static_cast<REAL4m>(Nb_Atof(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,antialias"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) arrow_config.antialias=atoi(cptr);

  // Pixel plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,status"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) pixel_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("pixel,colorcount"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) colorcount=atoi(cptr);
  if(colorcount<0) colorcount=pixel_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("pixel,colormap"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    pixel_config.colormap.Setup(colorcount,cptr);

  pixel_config.stipple="";
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,opaque"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    int opaque=atoi(cptr);
    if(!opaque) pixel_config.stipple="gray25";
  }

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,colorphase"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) phase=Nb_Atof(cptr);
  else                     phase=0.;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,colorinvert"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) invert=atoi(cptr);
  else                     invert=0;

  myMesh->ColorQuantityTransform(flipstr,colorquantity,phase,invert,
                                 pixel_config.colorquantity,
				 pixel_config.phase,pixel_config.invert);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,autosample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) pixel_config.autosample=atoi(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    pixel_config.subsample=static_cast<REAL4m>(Nb_Atof(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,size"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    pixel_config.size=static_cast<REAL4m>(Nb_Atof(cptr));

  // Misc.
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,drawboundary"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    myFrame.SetDrawBoundary(atoi(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,background"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) myFrame.SetBackgroundColor(cptr);

  // Get and set quantity type list
  Nb_DString* sp;
  Nb_List<Nb_DString> cqtypes;
  int cqtypec=myFrame.GetColorQuantityTypes(cqtypes);
  if (cqtypec>0) {
    char **cqtypev=new char*[cqtypec];
    int i;
    for(i=0, sp=cqtypes.GetFirst() ;
	sp!=NULL && i<cqtypec ; i++,sp=cqtypes.GetNext()) {
      cqtypev[i]=new char[sp->Length()+1];
      strcpy(cqtypev[i],sp->GetStr());
    }
    int itop=i;
    char *cqlist=Tcl_Merge(cqtypec,cqtypev);
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("quantitylist"),
		cqlist,TCL_GLOBAL_ONLY);
    Tcl_Free(cqlist);
    for(i=0;i<itop;i++) delete[] cqtypev[i];
    delete[] cqtypev;
  } else {
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("quantitylist"),
		ab1(""),TCL_GLOBAL_ONLY);
  }

  // Try to make selected color quantities valid
  int acq_match=0,pcq_match=0;
  const char* dfcq=NULL;
  for(sp=cqtypes.GetFirst() ; sp!=NULL ; sp=cqtypes.GetNext()) {
    if(strcmp(arrow_config.colorquantity,sp->GetStr())==0) acq_match=1;
    if(strcmp(pixel_config.colorquantity,sp->GetStr())==0) pcq_match=1;
    if(dfcq==NULL) dfcq=sp->GetStr();
  }
  if(!acq_match && dfcq!=NULL) {
    arrow_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }
  if(!pcq_match && dfcq!=NULL) {
    pixel_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }

  // Apply new configuration
  myFrame.SetPlotConfiguration(arrow_config,pixel_config);
  return TCL_OK;
}

int SetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),"SetFrameRotation must be called with"
	    " 1 argument: new_rotation_angle"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  myFrame.SetCoordinates(AngleToCoords(
                         static_cast<REAL4m>(Nb_Atof(argv[1]))));

  return TCL_OK;
}

int GetFrameBox(ClientData,Tcl_Interp *interp,
		 int argc,CONST84 char**)
{ // NOTE: The return value from this routine is a *6*-tuple,
  //  "xmin ymin zmin xmax ymax zmax".
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetFrameBox should be called with no arguments."
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_BoundingBox<REAL4> bb=myFrame.GetDisplayBox();
  Nb_Vec3<REAL4> minpt,maxpt;
  bb.GetExtremes(minpt,maxpt);
  if(bb.IsEmpty()) maxpt=minpt;  // Empty boxes may have maxpt<minpt.

  Oc_Snprintf(buf,sizeof(buf),"%g %g %g %g %g %g",
	  minpt.x,minpt.y,minpt.z,maxpt.x,maxpt.y,maxpt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetZoom(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),"GetZoom must be called with"
	    " 1 argument: name of zoom variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Oc_Snprintf(buf,sizeof(buf),"%.6g",myFrame.GetZoom());
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

int SetZoom(ClientData,Tcl_Interp *interp,
		  int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<2 || argc>3) {
    Oc_Snprintf(buf,sizeof(buf),"SetZoom must be called with"
	    " either 1 or 2 arguments: <new_zoom|new_width new_height>"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Change mesh scale in display list
  REAL4m newzoom;
  if(argc==2) newzoom=myFrame.SetZoom(REAL4m(Nb_Atof(argv[1])));
  else        newzoom=myFrame.SetZoom(REAL4m(Nb_Atof(argv[1])),
				      REAL4m(Nb_Atof(argv[2])));

  Oc_Snprintf(buf,sizeof(buf),"%g",newzoom);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int GetDataValueUnit(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),"GetDataValueUnit must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Oc_Snprintf(buf,sizeof(buf),"%.*s",MY_BUF_SIZE,myFrame.GetValueUnit());
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}

int GetDataValueScaling(ClientData,Tcl_Interp *interp,
			int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetDataValueScaling must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Oc_Snprintf(buf,sizeof(buf),"%.6g",myFrame.GetValueScaling());
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}


int SetDataValueScaling(ClientData,Tcl_Interp *interp,
		      int argc,CONST84 char** argv)
{
  // Adjust data value scaling
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
            "SetDataValueScaling must be called with"
	    " 1 argument: <new_scale>"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL8m newscale=myFrame.SetValueScaling(Nb_Atof(argv[1]));
  Oc_Snprintf(buf,sizeof(buf),"%g",newscale);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int WriteMeshUsingDeprecatedVIOFormat(ClientData,Tcl_Interp *interp,
	      int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc != 2) {
    Oc_Snprintf(buf,sizeof(buf),
            "WriteMeshDeprecatedVIOFormat must be called with"
	    " 1 argument: filename"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  // Open file
  FILE *fptr;
  if (argv[1][0] == '\0') {
    fptr=stdout;
  } else {
    fptr=Nb_FOpen(argv[1],"wb");
  }
  if(fptr==NULL) {
    Oc_Snprintf(buf,sizeof(buf),
	    "Error in %s: Unable to open file %s for output",
	    argv[0], argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    errcode=TCL_ERROR;
  }
  else {
    Vf_VioFileOutput vfo;
    if(vfo.WriteMesh(myMesh,fptr,NULL)!=0) {
      Oc_Snprintf(buf,sizeof(buf),"WriteMeshDeprecatedVIOFormat error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
    }
    fflush(fptr);
    if (argv[1][0] != '\0') {
      fclose(fptr);
    }
  }

  return errcode;
}

int WriteMesh(ClientData,Tcl_Interp *interp,
	      int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<4 || argc>6) {
    Oc_Snprintf(buf,sizeof(buf),"WriteMesh must be called with"
	    " 3-5 arguments: filename <text|binary4|binary8>"
	    " <regular|irregular> [title] [description]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  Vf_OvfDataStyle ods=vf_oascii;
  if(strcmp("binary4",argv[2])==0)      ods=vf_obin4;
  else if(strcmp("binary8",argv[2])==0) ods=vf_obin8;

  BOOL force_irreg=0;
  if(strcmp("irregular",argv[3])==0) force_irreg=1;

  const char *title="",*desc="";
  if(argc>4) title=argv[4];
  if(argc>5) desc=argv[5];

  // Open file
  FILE *fptr;
  if (argv[1][0] == '\0') {
    fptr=stdout;
  } else {
    fptr=Nb_FOpen(argv[1],"wb");
  }
  if(fptr==NULL) {
    Oc_Snprintf(buf,sizeof(buf),
            "Error in %s: Unable to open file %s for output", 
	    argv[0], argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    errcode=TCL_ERROR;
  }
  else {
    Vf_OvfFileOutput ofo;
    if(ofo.WriteMesh(myMesh,fptr,ods,force_irreg,title,desc)!=0) {
      Oc_Snprintf(buf,sizeof(buf),"WriteMesh error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
    }
    fflush(fptr);
    if (argv[1][0] != '\0') {
      fclose(fptr);
    }
  }

  return errcode;
}

int
WriteMeshAverages(ClientData, Tcl_Interp *interp, int argc, CONST84 char** argv)
{
  Nb_DString colhdrfmt,linefmt;
// #define NUMFMT "%- #24.17g"
#define NUMFMT "%- 10.15g"
  Tcl_Channel channel;
  int mode;

  Tcl_ResetResult(interp);
  if(argc<5 || argc>6) {
    Oc_Snprintf(buf,sizeof(buf),
            "WriteMeshAverages must be called with 5 or 6 arguments:"
	    " an output channel, average type"
	    " (one of space|plane|line|point),"
	    " axis selection (x|y|z),"
	    " range selection (6-tuple),"
	    " and an optional descriptor string"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  if ((channel = Tcl_GetChannel(interp, argv[1], &mode)) == NULL) {
    return TCL_ERROR;
  }
  if (!(mode & TCL_WRITABLE)) {
    Oc_Snprintf(buf,sizeof(buf),"%s is not a writable channel",argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Averaging type
  enum { point, line, plane, space } avetype = space;
  if(strcmp(argv[2],"point")==0) {
    avetype = point;
  } else if(strcmp(argv[2],"line")==0) {
    avetype = line;
  } else if(strcmp(argv[2],"plane")==0) {
    avetype = plane;
  } else if(strcmp(argv[2],"space")!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid type string: %s;"
	    " Should be one of point, line, plane or space",
	    argv[2]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Average type axis spec
  if(argv[3][1]!='\0' ||
     (argv[3][0]!='x' && argv[3][0]!='y' && argv[3][0]!='z')) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid axis spec: %s;"
	    " Should be x, y, or z",
	    argv[3]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  char axis = argv[3][0];

    // Is this a rectangular mesh?
  if(strcmp(myMesh->GetMeshType(),"Vf_GridVec3f")!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid mesh type: %s\n",
		myMesh->GetMeshType());
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Downcast
  Vf_GridVec3f& grid = *((Vf_GridVec3f *)myMesh);
  REAL8m scale=grid.GetValueMultiplier()
    * grid.GetDisplayValueScale();

  // Check dimensions
  int i,j,k,isize,jsize,ksize;
  grid.GetDimens(isize,jsize,ksize);
  if(isize<1 || jsize<1 || ksize<1) {
    Oc_Snprintf(buf,sizeof(buf),"Bad mesh dimensions: %d %d %d",
		isize,jsize,ksize);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  double mult=0.;
  Nb_Vec3<REAL8> base=grid.GetBasePoint();
  Nb_Vec3<REAL8> step=grid.GetGridStep();

  // Range select
  int range_arr_size;
  CONST84 char ** range_arr;
  if (Tcl_SplitList(interp,argv[4],&range_arr_size,&range_arr)!=TCL_OK) {
    return TCL_ERROR;
  }
  if(range_arr_size!=6) {
    Tcl_Free((char *)range_arr);
    Oc_Snprintf(buf,sizeof(buf),"Range list has %d != 6 elements",
		range_arr_size);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  REAL8m xmin = base.x - step.x/2.;
  REAL8m ymin = base.y - step.y/2.;
  REAL8m zmin = base.z - step.z/2.;
  REAL8m xmax = base.x + (isize-0.5)*step.x;
  REAL8m ymax = base.y + (jsize-0.5)*step.y;
  REAL8m zmax = base.z + (ksize-0.5)*step.z;
  if(xmin>xmax) { REAL8m temp=xmin; xmin=xmax; xmax=temp; }
  if(ymin>ymax) { REAL8m temp=ymin; ymin=ymax; ymax=temp; }
  if(zmin>zmax) { REAL8m temp=zmin; zmin=zmax; zmax=temp; }
  if(strcmp("-",range_arr[0])!=0) {
    REAL8m temp = Nb_Atof(range_arr[0]);
    if(temp>xmin) xmin=temp;
  }
  if(strcmp("-",range_arr[1])!=0) {
    REAL8m temp = Nb_Atof(range_arr[1]);
    if(temp>ymin) ymin=temp;
  }
  if(strcmp("-",range_arr[2])!=0) {
    REAL8m temp = Nb_Atof(range_arr[2]);
    if(temp>zmin) zmin=temp;
  }
  if(strcmp("-",range_arr[3])!=0) {
    REAL8m temp = Nb_Atof(range_arr[3]);
    if(temp<xmax) xmax=temp;
  }
  if(strcmp("-",range_arr[4])!=0) {
    REAL8m temp = Nb_Atof(range_arr[4]);
    if(temp<ymax) ymax=temp;
  }
  if(strcmp("-",range_arr[5])!=0) {
    REAL8m temp = Nb_Atof(range_arr[5]);
    if(temp<zmax) zmax=temp;
  }
  Tcl_Free((char *)range_arr);

  if(xmin>xmax || ymin>ymax || zmin>zmax) {
    // Nothing to do.
    Oc_Snprintf(buf,sizeof(buf),"## Empty range\n");
    Tcl_Write(channel, buf, int(strlen(buf)));
    return TCL_OK;
    /// NOTE: Checking here on the ?min/?max float values
    ///  protects against some overflow conditions on the
    ///  parallel integer constructions below
  }
  int imin=0,imax=isize;
  if(step.x>0) {
    imin = int(ceil((xmin - base.x)/step.x));
    imax = int(floor((xmax - base.x)/step.x));
  } else if(step.x<0) {
    imin = int(ceil((xmax - base.x)/step.x));
    imax = int(floor((xmin - base.x)/step.x));
  }
  ++imax;
  if(imin<0)     imin = 0;
  if(imax>isize) imax = isize;
  int icount = imax - imin;

  int jmin=0,jmax=jsize;
  if(step.y>0) {
    jmin = int(ceil((ymin - base.y)/step.y));
    jmax = int(floor((ymax - base.y)/step.y));
  } else if(step.y<0) {
    jmin = int(ceil((ymax - base.y)/step.y));
    jmax = int(floor((ymin - base.y)/step.y));
  }
  ++jmax;
  if(jmin<0)     jmin = 0;
  if(jmax>jsize) jmax = jsize;
  int jcount = jmax - jmin;

  int kmin=0,kmax=ksize;
  if(step.z>0) {
    kmin = int(ceil((zmin - base.z)/step.z));
    kmax = int(floor((zmax - base.z)/step.z));
  } else if(step.z<0) {
    kmin = int(ceil((zmax - base.z)/step.z));
    kmax = int(floor((zmin - base.z)/step.z));
  }
  ++kmax;
  if(kmin<0)     kmin = 0;
  if(kmax>ksize) kmax = ksize;
  int kcount = kmax - kmin;

  if(icount<1 || jcount<1 || kcount<1) {
    // Nothing to do
    Oc_Snprintf(buf,sizeof(buf),"## Empty range\n");
    Tcl_Write(channel, buf, int(strlen(buf)));
    return TCL_OK;
  }

  // Print ODT file header
  Oc_Snprintf(buf,sizeof(buf),"# ODT 1.0\n");
  Tcl_Write(channel, buf, int(strlen(buf)));
  if(argc==6) {
    Oc_Snprintf(buf,sizeof(buf),"## Description: %s\n",argv[5]);
    Tcl_Write(channel, buf, int(strlen(buf)));
  }

  unsigned int volume_point_count = icount*jcount*kcount;
  Oc_Snprintf(buf,sizeof(buf),
	  "## Active volume: (%.15g,%.15g,%.15g) x (%.15g,%.15g,%.15g)\n"
	  "## Cell size: %.15g x %.15g x %.15g\n"
	  "## Cells in active volume: %d\n",
	  xmin,ymin,zmin,xmax,ymax,zmax,
	  fabs(step.x),fabs(step.y),fabs(step.z),
	  volume_point_count);
  Tcl_Write(channel, buf, int(strlen(buf)));

  Oc_Snprintf(buf,sizeof(buf),"#\n# Table Start\n");
  Tcl_Write(channel, buf, int(strlen(buf)));

  Nb_Vec3<REAL8m> m;

  if(avetype==space) {
    // Volume average
    colhdrfmt = Nb_DString("# Columns: %-25s  %-25s %-25s\n")
      + Nb_DString("# Units: {} {} {}\n");
    linefmt = Nb_DString(NUMFMT)
      + Nb_DString("  ") + Nb_DString(NUMFMT)
      + Nb_DString("  ") + Nb_DString(NUMFMT)
      + Nb_DString("\n");
    Oc_Snprintf(buf,sizeof(buf),
		"# Title: Average across active volume\n");
    Tcl_Write(channel, buf, int(strlen(buf)));
    Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
	    "    m_x","    m_y","    m_z");
    Tcl_Write(channel, buf, int(strlen(buf)));
    m.Set(0.,0.,0.);
    for(k=kmin;k<kmax;k++) for(j=jmin;j<jmax;j++) for(i=imin;i<imax;i++) {
      m+=grid(i,j,k);
    }
    mult=(scale/double(volume_point_count));
    m*=mult;
    Oc_Snprintf(buf,sizeof(buf),linefmt,m.x,m.y,m.z);
    Tcl_Write(channel, buf, int(strlen(buf)));
  } else {
    switch(avetype) {
    case plane:
      colhdrfmt = Nb_DString("# Columns: %-25s  %-25s %-25s %-25s\n")
	+ Nb_DString("# Units: m {} {} {}\n");
      linefmt = Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("\n");
      if(axis == 'x') {
	// Collect and print yz-averaged data along x-axis
	Oc_Snprintf(buf,sizeof(buf),
		"# Title: Averages across x-axis (%d points each)\n",
		jcount*kcount);
	Tcl_Write(channel, buf, int(strlen(buf)));
	Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
		"     x","    m_x","    m_y","    m_z");
	Tcl_Write(channel, buf, int(strlen(buf)));
	REAL8m mult=(scale/double(jcount*kcount));
	for(i=imin;i<imax;i++) {
	  m.Set(0.,0.,0.);
	  for(k=kmin;k<kmax;k++) for(j=jmin;j<jmax;j++) {
	    m+=grid(i,j,k);
	  }
	  m*=mult;
	  Oc_Snprintf(buf,sizeof(buf),linefmt,
		  base.x+double(i)*step.x,m.x,m.y,m.z);
	  Tcl_Write(channel, buf, int(strlen(buf)));
	}
      } else if(axis == 'y') {
	// Collect and print xz-averaged data along y-axis
	Oc_Snprintf(buf,sizeof(buf),
		"# Title: Averages across y-axis (%d points each)\n",
		icount*kcount);
	Tcl_Write(channel, buf, int(strlen(buf)));
	Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
		"     y","    m_x","    m_y","    m_z");
	Tcl_Write(channel, buf, int(strlen(buf)));
	REAL8m mult=(scale/double(icount*kcount));
	for(j=jmin;j<jmax;j++) {
	  m.Set(0.,0.,0.);
	  for(k=kmin;k<kmax;k++) for(i=imin;i<imax;i++) {
	    m+=grid(i,j,k);
	  }
	  m*=mult;
	  Oc_Snprintf(buf,sizeof(buf),linefmt,
		  base.y+double(j)*step.y,m.x,m.y,m.z);
	  Tcl_Write(channel, buf, int(strlen(buf)));
	}
      } else if(axis == 'z') {
	// Collect and print xy-averaged data along z-axis
	Oc_Snprintf(buf,sizeof(buf),
		"# Title: Averages across z-axis (%d points each)\n",
		icount*jcount);
	Tcl_Write(channel, buf, int(strlen(buf)));
	Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
		"     z","    m_x","    m_y","    m_z");
	Tcl_Write(channel, buf, int(strlen(buf)));
	REAL8m mult=(scale/double(icount*jcount));
	for(k=kmin;k<kmax;k++) {
	  m.Set(0.,0.,0.);
	  for(j=jmin;j<jmax;j++) for(i=imin;i<imax;i++) {
	    m+=grid(i,j,k);
	  }
	  m*=mult;
	  Oc_Snprintf(buf,sizeof(buf),linefmt,
		  base.z+double(k)*step.z,m.x,m.y,m.z);
	  Tcl_Write(channel, buf, int(strlen(buf)));
	}
      }
      break;
    case line:
      colhdrfmt = Nb_DString("# Columns: %-25s  %-25s %-25s %-25s %-25s\n")
	+ Nb_DString("# Units: m m {} {} {}\n");
      linefmt = Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("\n");
      if(axis == 'x') {
	// Print data averages along lines parallel to x-axis
	Oc_Snprintf(buf,sizeof(buf),
		"# Title: Averages parallel to x-axis (%d points each)\n",
		icount);
	Tcl_Write(channel, buf, int(strlen(buf)));
	Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
		"     y","    z","    m_x","    m_y","    m_z");
	Tcl_Write(channel, buf, int(strlen(buf)));
	REAL8m mult=(scale/double(icount));
	for(k=kmin;k<kmax;k++) for(j=jmin;j<jmax;j++) {
	  m.Set(0.,0.,0.);
	  for(i=imin;i<imax;i++) {
	    m+=grid(i,j,k);
	  }
	  m*=mult;
	  Oc_Snprintf(buf,sizeof(buf),linefmt,
		  base.y+double(j)*step.y,
		  base.z+double(k)*step.z,
		  m.x,m.y,m.z);
	  Tcl_Write(channel, buf, int(strlen(buf)));
	}
      } else if(axis == 'y') {
	// Print data averages along lines parallel to y-axis
	Oc_Snprintf(buf,sizeof(buf),
		"# Title: Averages parallel to y-axis (%d points each)\n",
		jcount);
	Tcl_Write(channel, buf, int(strlen(buf)));
	Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
		"     x","    z","    m_x","    m_y","    m_z");
	Tcl_Write(channel, buf, int(strlen(buf)));
	REAL8m mult=(scale/double(jcount));
	for(k=kmin;k<kmax;k++) for(i=imin;i<imax;i++) {
	  m.Set(0.,0.,0.);
	  for(j=jmin;j<jmax;j++) {
	    m+=grid(i,j,k);
	  }
	  m*=mult;
	  Oc_Snprintf(buf,sizeof(buf),linefmt,
		  base.x+double(i)*step.x,
		  base.z+double(k)*step.z,
		  m.x,m.y,m.z);
	  Tcl_Write(channel, buf, int(strlen(buf)));
	}
      } else if(axis == 'z') {
	// Print data averages along lines parallel to z-axis
	Oc_Snprintf(buf,sizeof(buf),
		"# Title: Averages parallel to z-axis (%d points each)\n",
		kcount);
	Tcl_Write(channel, buf, int(strlen(buf)));
	Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
		"     x","    y","    m_x","    m_y","    m_z");
	Tcl_Write(channel, buf, int(strlen(buf)));
	REAL8m mult=(scale/double(kcount));
	for(j=jmin;j<jmax;j++) for(i=imin;i<imax;i++) {
	  m.Set(0.,0.,0.);
	  for(k=kmin;k<kmax;k++) {
	    m+=grid(i,j,k);
	  }
	  m*=mult;
	  Oc_Snprintf(buf,sizeof(buf),linefmt,
		  base.x+double(i)*step.x,
		  base.y+double(j)*step.y,
		  m.x,m.y,m.z);
	  Tcl_Write(channel, buf, int(strlen(buf)));
	}
      }
      break;
    case point:
      colhdrfmt = Nb_DString("# Title: Points in specified volume\n")
	+ Nb_DString("# Columns: %-25s  %-25s %-25s %-25s %-25s %-25s\n")
	+ Nb_DString("# Units: m m {} {} {}\n");
      linefmt = Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("  ") + Nb_DString(NUMFMT)
	+ Nb_DString("\n");
      // Print data points inside specified volume
      Oc_Snprintf(buf,sizeof(buf),colhdrfmt,
	      "     x","     y","    z",
	      "    m_x","    m_y","    m_z");
      Tcl_Write(channel, buf, int(strlen(buf)));
      for(k=kmin;k<kmax;k++) for(j=jmin;j<jmax;j++)
	for(i=imin;i<imax;i++) {
	  m=grid(i,j,k);
	  m*=scale;
	  Oc_Snprintf(buf,sizeof(buf),linefmt,
		  base.x+double(i)*step.x,
		  base.y+double(j)*step.y,
		  base.z+double(k)*step.z,
		  m.x,m.y,m.z);
	  Tcl_Write(channel, buf, int(strlen(buf)));
	}
      break;
    case space:
      Oc_Snprintf(buf,sizeof(buf),"Programming error: 'space' avetype"
		  " not properly handled");
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
      // Implicit "break"
    }
  }
  // Print ODT file trailer
  Oc_Snprintf(buf,sizeof(buf),"# Table End\n");
  Tcl_Write(channel, buf, int(strlen(buf)));
  Tcl_Flush(channel);
  return TCL_OK;
}

int
BitmapCmd(ClientData clientData, Tcl_Interp *interp, int argc, CONST84 char** argv)
{
  char c;
  const char* cptr;
  int margin, width, height, crop, isNew, mode;
  int xmin, ymin, xmax, ymax, xmarginadj, ymarginadj;
  int draw_boundary;
  REAL8m boundary_width;
  REAL8m mat_width=0.0;  // Mat width, in pixels
  Nb_DString mat_color = "0xFFFFFF";  // Default color is white
  OommfBitmap *bitmapPtr;
  Oc_AutoBuf ab1, ab2;
  Tcl_HashTable *tablePtr = (Tcl_HashTable *)clientData;
  Tcl_HashEntry *entryPtr;
  Tcl_Channel channel;
  
  CHAR cmdbuf[MY_BUF_SIZE];

  if (argc < 2) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s option bitmapName"
		" ?arg ...?\"", argv[0]);
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }
  c = argv[1][0];
  if ((c == 'C') && (strcmp(argv[1], "Create") == 0)) {
    if (argc != 3) {
      Oc_Snprintf(buf,sizeof(buf),
		  "wrong # args: should be \"%s Create bitmapName\"",
		  argv[0]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    entryPtr = Tcl_CreateHashEntry(tablePtr, argv[2], &isNew);
    if (!isNew) {
      Oc_Snprintf(buf,sizeof(buf),
		  "bitmap \"%s\" already exists", argv[2]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }

    // Create a new OommfBitmap from the global DisplayFrame
    bitmapPtr = new OommfBitmap;

    margin = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			      ab2("misc,margin"), TCL_GLOBAL_ONLY));
    width = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			     ab2("misc,width"), TCL_GLOBAL_ONLY));
    height = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			      ab2("misc,height"), TCL_GLOBAL_ONLY));
    crop = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			    ab2("misc,crop"), TCL_GLOBAL_ONLY));

    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,drawboundary"),TCL_GLOBAL_ONLY);
    draw_boundary=1;
    if(!Nb_StrIsSpace(cptr)) draw_boundary = atoi(cptr);
    if(draw_boundary) myFrame.SetDrawBoundary(1);
    else              myFrame.SetDrawBoundary(0);

    BOOL bw_error;
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,boundarywidth"),TCL_GLOBAL_ONLY);
    boundary_width = Nb_Atof(cptr,bw_error);
    if(!bw_error) myFrame.SetBoundaryWidth(boundary_width);

    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,boundarycolor"),TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) myFrame.SetBoundaryColor(cptr);

    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,boundarypos"),TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      if(Nb_StrCaseCmp(cptr,"back")==0) {
	myFrame.SetBoundaryOnTop(0);
      } else if(Nb_StrCaseCmp(cptr,"front")==0) {
	myFrame.SetBoundaryOnTop(1);
      } else {
	Oc_Snprintf(buf,sizeof(buf),
		    "Invalid plot_config(misc,boundarypos) value:"
		    " \"%s\"; should be either \"front\" or \"back\"",
		    cptr);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
    }

    BOOL mw_error;
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,matwidth"),TCL_GLOBAL_ONLY);
    REAL8m mw_temp = Nb_Atof(cptr,mw_error);
    if(!mw_error) mat_width = mw_temp; // Default is 0.0
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,matcolor"),TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      mat_color = cptr;   // Default is OxFFFFFF (white)
    }

    // Mesh box
    Nb_Vec3<REAL8> minpt, maxpt;
    Nb_BoundingBox<REAL8> meshbox;
    myMesh->GetPreciseRange(meshbox);
    meshbox.GetExtremes(minpt,maxpt);

    // Set up slice selection
    char viewaxis = 'z';
    char viewdir[3] = "+z";
    BOOL negative_viewdir = 0;
    BOOL center_error,arrowspan_error,pixelspan_error;
    REAL8m center,arrowspan,pixelspan;
    cptr = Tcl_GetVar2(interp, ab1("plot_config"),
		       ab2("viewaxis"), TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      size_t strsize = strlen(cptr);
      if(strsize==1) {
	viewaxis = cptr[0];
	viewdir[0] = '+'; viewdir[1]=viewaxis; viewdir[2]='\0';
      } else if(strsize==2) {
	viewaxis = cptr[1];
	strcpy(viewdir,cptr);
	if(cptr[0]=='-') {
	  negative_viewdir = 1; // Viewing from the backside
	}
      } else {
	Oc_Snprintf(buf,sizeof(buf),
		    "Invalid plot_config(viewaxis) string : \"%s\"",
		    cptr);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
    }


    BOOL centerpt_set = 0;
    Nb_Vec3<REAL8m> centerpt;
    cptr = Tcl_GetVar2(interp, ab1("plot_config"),
		       ab2("misc,centerpt"), TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      // Centerpt set in plot_config.  Read this it in and convert
      // to viewaxis coords
      Oc_Snprintf(cmdbuf,sizeof(cmdbuf),"ApplyAxisTransform +z %s %s",
		  viewdir,cptr);
      Tcl_SavedResult saved;
      Tcl_SaveResult(interp, &saved);
      int error_code=Tcl_Eval(interp,cmdbuf);
      if(error_code!=TCL_OK) {
	Oc_Snprintf(buf,sizeof(buf),
		    "Error processing plot_config(misc,centerpt) string :"
		    " \"%s\"",cptr);
	return error_code;
      }
      double x,y,z;
      if(sscanf(Tcl_GetStringResult(interp),"%lg%lg%lg",&x,&y,&z)!=3) {
	// This shouldn't happen!
	Oc_Snprintf(buf,sizeof(buf),
		    "LOGIC ERROR! While processing "
		    "plot_config(misc,centerpt) string : \"%s\"",cptr);
	return TCL_ERROR;
      }
      centerpt.Set(x,y,z);
      centerpt_set = 1;
      Tcl_RestoreResult(interp, &saved);
    }

    if(viewaxis!='x' && viewaxis!='y' && viewaxis!='z') {
      Oc_Snprintf(buf,sizeof(buf),
		  "Invalid viewaxis detected: %c; Check setting of"
		  " plot_config(viewaxis)",viewaxis);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }

    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("viewaxis,center"),TCL_GLOBAL_ONLY);
    center = Nb_Atof(cptr,center_error);
    if(negative_viewdir) center *= -1; // Adjust coords.

    Oc_Snprintf(buf,sizeof(buf),"viewaxis,%carrowspan",viewaxis);
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),buf,TCL_GLOBAL_ONLY);
    arrowspan = Nb_Atof(cptr,arrowspan_error);
    
    Oc_Snprintf(buf,sizeof(buf),"viewaxis,%cpixelspan",viewaxis);
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),buf,TCL_GLOBAL_ONLY);
    pixelspan = Nb_Atof(cptr,pixelspan_error);

    // Mesh is always rotated so slice is along 'z' axis wrt mesh
    // coordinates.
    if(center_error)    {
      if(centerpt_set) center = centerpt.z;
      else             center = (minpt.z+maxpt.z)/2.0;
    }
    if(arrowspan_error) arrowspan = (maxpt.z-minpt.z);
    if(pixelspan_error) pixelspan = (maxpt.z-minpt.z);

    // Calculate bounding boxes, in mesh coords
    Nb_Vec3<REAL4> temp_min,temp_max;
    Convert(minpt,temp_min);  Convert(maxpt,temp_max);
    temp_min.z = static_cast<REAL4>(center - arrowspan/2);
    temp_max.z = static_cast<REAL4>(center + arrowspan/2);
    Nb_BoundingBox<REAL4> arrow_box(temp_min,temp_max);
    temp_min.z = static_cast<REAL4>(center - pixelspan/2);
    temp_max.z = static_cast<REAL4>(center + pixelspan/2);
    Nb_BoundingBox<REAL4> pixel_box(temp_min,temp_max);

    // Setup bitmap
    Nb_Vec3<REAL4> disp_minpt,disp_maxpt;
    Nb_BoundingBox<REAL4> bbox;
    bbox = myFrame.GetDisplayBox();
    bbox.GetExtremes(disp_minpt, disp_maxpt);
    xmin = int(floor(disp_minpt.x)) - margin;
    ymin = int(floor(disp_minpt.y)) - margin;
    xmax = int(ceil(disp_maxpt.x)) + margin;
    ymax = int(ceil(disp_maxpt.y)) + margin;
    xmarginadj = (xmax - xmin + 1) - width;
    ymarginadj = (ymax - ymin + 1) - height;
    if(xmarginadj>0) {
      // Resulting size is bigger than requested. Trim.
      xmin += int(floor(xmarginadj/2.));
      xmax -= int(ceil(xmarginadj/2.));
    }
    if(ymarginadj>0) {
      // Resulting size is bigger than requested. Trim.
      ymin += int(floor(ymarginadj/2.));
      ymax -= int(ceil(ymarginadj/2.));
    }
    if(!crop) {
      // No-crop request, so make bitmap exact size requested
      xmax = xmin + width - 1;
      ymax = ymin + height - 1;
    }
    // If centerpt is specified, then pan bitmap window
    if(centerpt_set) {
      Nb_Vec3<REAL4> view_center,display_center;
      Convert(centerpt,view_center);
      display_center.x = static_cast<REAL4>((minpt.x+maxpt.x)/2.);
      display_center.y = static_cast<REAL4>((minpt.y+maxpt.y)/2.);
      display_center.z = static_cast<REAL4>((minpt.z+maxpt.z)/2.);
      /// Note: display_center.z is not actually used
      myFrame.CoordinatePointTransform(CS_CALCULATION_STANDARD,
				       myFrame.GetCoordinates(),
				       view_center);
      myFrame.CoordinatePointTransform(CS_CALCULATION_STANDARD,
				       myFrame.GetCoordinates(),
				       display_center);
      int xadj = int(OC_ROUND(view_center.x-display_center.x));
      int yadj = int(OC_ROUND(view_center.y-display_center.y));
      xmin += xadj;   xmax += xadj;
      ymin += yadj;   ymax += yadj;
    }
    bitmapPtr->Setup(xmin,ymin,xmax,ymax);

    // Fill bitmap
    myFrame.FillBitmap(arrow_box,pixel_box,negative_viewdir,*bitmapPtr);

    // Add mat, if requested
    if(mat_width>0) {
      OommfPackedRGB color(mat_color.GetStr());
      bitmapPtr->AddMat(mat_width,color);
    }

    Tcl_SetHashValue(entryPtr, (ClientData)bitmapPtr);
    return TCL_OK;
  }
  if ((c == 'D') && (strcmp(argv[1], "Delete") == 0)) {
    if (argc != 3) {
      Oc_Snprintf(buf,sizeof(buf),
		  "wrong # args: should be \"%s Delete bitmapName\"",
		  argv[0]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    entryPtr = Tcl_FindHashEntry(tablePtr, argv[2]);
    if (entryPtr == NULL) {
      Oc_Snprintf(buf,sizeof(buf), "can't find bitmap \"%s\"",
		  argv[2]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    delete (OommfBitmap *) Tcl_GetHashValue(entryPtr);
    Tcl_DeleteHashEntry(entryPtr);
    return TCL_OK;
  }
  if ((c == 'W') && (strcmp(argv[1], "Write") == 0)) {
    if (argc != 5) {
      Oc_Snprintf(buf,sizeof(buf), "wrong # args: should be "
		  "\"%s Write bitmapName format channel\"", argv[0]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    if ((channel = Tcl_GetChannel(interp, argv[4], &mode)) == NULL) {
      return TCL_ERROR;
    }
    if (!(mode & TCL_WRITABLE)) {
      Oc_Snprintf(buf,sizeof(buf),
		  "%s is not a writable channel", argv[4]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    entryPtr = Tcl_FindHashEntry(tablePtr, argv[2]);
    if (entryPtr == NULL) {
      Oc_Snprintf(buf,sizeof(buf),
		  "can't find bitmap \"%s\"", argv[2]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    bitmapPtr = (OommfBitmap *) Tcl_GetHashValue(entryPtr);
    if ( strcmp("P3", argv[3]) == 0) {
      bitmapPtr->WritePPMChannel(channel, 3);
      return TCL_OK;
    }
    if ( strcmp("P6", argv[3]) == 0) {
      bitmapPtr->WritePPMChannel(channel, 6);
      return TCL_OK;
    }
    if ( strcmp("B24", argv[3]) == 0) {
      bitmapPtr->WriteBMPChannel(channel, 24);
      return TCL_OK;
    }
    Oc_Snprintf(buf,sizeof(buf),
		"ERROR: Unknown file format request: %s\n",argv[3]);
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),
	      "bad option \"%s\": must be Create, Delete, or Write",
	      argv[1]);
  Tcl_AppendResult(interp, buf, (char *) NULL);
  return TCL_ERROR;
}

void
BitmapDeleteProc(ClientData clientData)
{
    Tcl_HashTable *tablePtr = (Tcl_HashTable *) clientData;
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchId;
    OommfBitmap *bitmapPtr;

    for (entryPtr = Tcl_FirstHashEntry(tablePtr, &searchId);
	    entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&searchId)) {
	bitmapPtr = (OommfBitmap *) Tcl_GetHashValue(entryPtr);
	delete bitmapPtr;
    }
    Tcl_DeleteHashTable(tablePtr);
    delete tablePtr;
}

static void
CleanupMesh(ClientData)
{
  for(int i=0;i<MY_MESH_ARRAY_SIZE;i++) {
    if (myMeshArray[i]!=NULL) {
      delete myMeshArray[i];
      myMeshArray[i]=NULL;  // Safety
    }
  }
  myMesh=NULL;     // Safety
  activeMeshId=0;  // Safety
}

int Mmdispcmds_Init(Tcl_Interp *interp)
{
#define RETURN_TCL_ERROR                                                \
    Tcl_AddErrorInfo(interp, ab1("\n    (in Mmdispcmds_Init())"));      \
    return TCL_ERROR

  Oc_AutoBuf ab1,ab2;
  Tcl_HashTable *bitmapTablePtr;

  // Verify run-time and compile-time versions of required support
  // extensions match.
//  if (Tcl_PkgRequire(interp, ab1("Tcl"), ab2(TCL_VERSION), 1) == NULL) {
//    Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds "), NULL);
//    Tcl_AppendResult(interp, ab1(MMDISPCMDS_VERSION), NULL);
//    Tcl_AppendResult(interp, ab1(" needs Tcl"), NULL);
//    Tcl_AppendResult(interp, ab1(TCL_VERSION), NULL);
//    Tcl_AppendResult(interp, ab1(")"), NULL);
//  }

/*
 * Need at least Oc 1.1.1.0 to get Oc_Config's compiletime_* features
 */
    if (Tcl_PkgRequire(interp, ab1("Oc"), ab2("1.1.1.0"), 0) == NULL) {
        Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds " MMDISPCMDS_VERSION
                " needs Oc 1.1.1.0)"), NULL);
        RETURN_TCL_ERROR;
    }

/*
 * Need at least Vf 1.2.0.2 to get Vf_Mesh::GetValueMagSpan()
 */
    if (Tcl_PkgRequire(interp, ab1("Vf"), ab2("1.2.0.2"), 0) == NULL) {
        Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds " MMDISPCMDS_VERSION
                " needs Vf 1.2.0.2)"), NULL);
        RETURN_TCL_ERROR;
    }
  
  // Record compile and run-time patch levels

  // Initialize C++ data structures
  for(int id=0;id<MY_MESH_ARRAY_SIZE;id++) {
    myMeshArray[id] = new Vf_EmptyMesh();
  }
  activeMeshId=0;
  myMesh=myMeshArray[activeMeshId];
  Tcl_CreateExitHandler((Tcl_ExitProc *) CleanupMesh, 0);
  myFrame.SetTclInterp(interp);
  myFrame.SetMesh(myMesh);

  // Register C/C++ routines with interpreter
  Oc_RegisterCommand(interp,"ChangeMesh",ChangeMesh);
  Oc_RegisterCommand(interp,"CopyMesh",CopyMesh);
  Oc_RegisterCommand(interp,"DifferenceMesh",DifferenceMesh);
  Oc_RegisterCommand(interp,"DrawFrame",DrawFrame);
  Oc_RegisterCommand(interp,"FreeMesh",FreeMesh);
  Oc_RegisterCommand(interp,"GetDataValueUnit",GetDataValueUnit);
  Oc_RegisterCommand(interp,"GetDataValueScaling",GetDataValueScaling);
  Oc_RegisterCommand(interp,"GetFrameBox",GetFrameBox);
  Oc_RegisterCommand(interp,"GetFrameRotation",GetFrameRotation);
  Oc_RegisterCommand(interp,"GetZoom",GetZoom);
  Oc_RegisterCommand(interp,"GetMeshCellSize",GetMeshCellSize);
  Oc_RegisterCommand(interp,"GetMeshDescription",GetMeshDescription);
  Oc_RegisterCommand(interp,"GetMeshIncrement",GetMeshIncrement);
  Oc_RegisterCommand(interp,"GetMeshName",GetMeshName);
  Oc_RegisterCommand(interp,"GetMeshStructureInfo",GetMeshStructureInfo);
  Oc_RegisterCommand(interp,"GetMeshTitle",GetMeshTitle);
  Oc_RegisterCommand(interp,"GetMeshType",GetMeshType);
  Oc_RegisterCommand(interp,"GetMeshSpatialUnitString",
		     GetMeshSpatialUnitString);
  Oc_RegisterCommand(interp,"GetMeshRange",GetMeshRange);
  Oc_RegisterCommand(interp,"GetMeshZRange",GetMeshZRange);
  Oc_RegisterCommand(interp,"GetVecColor",GetVecColor);
  Oc_RegisterCommand(interp,"GetZsliceCount",GetZsliceCount);
  Oc_RegisterCommand(interp,"GetZsliceLevels",GetZsliceLevels);
  Oc_RegisterCommand(interp,"IsRegularMesh",IsRegularMesh);
  Oc_RegisterCommand(interp,"ReportActiveMesh",ReportActiveMesh);
  Oc_RegisterCommand(interp,"SelectActiveMesh",SelectActiveMesh);
  Oc_RegisterCommand(interp,"SetDataValueScaling",SetDataValueScaling);
  Oc_RegisterCommand(interp,"SetFrameRotation",SetFrameRotation);
  Oc_RegisterCommand(interp,"SetZoom",SetZoom);
  Oc_RegisterCommand(interp,"SetMeshTitle",SetMeshTitle);
  Oc_RegisterCommand(interp,"UpdatePlotConfiguration",
                     UpdatePlotConfiguration);
  Oc_RegisterCommand(interp,"WriteMesh",WriteMesh);
  Oc_RegisterCommand(interp,"WriteMeshUsingDeprecatedVIOFormat",
		     WriteMeshUsingDeprecatedVIOFormat);
  Oc_RegisterCommand(interp,"WriteMeshAverages",WriteMeshAverages);

  bitmapTablePtr = new Tcl_HashTable;
  Tcl_InitHashTable(bitmapTablePtr, TCL_STRING_KEYS);
  Tcl_CreateCommand(interp, ab1("Bitmap"), BitmapCmd,
      (ClientData) bitmapTablePtr, BitmapDeleteProc);

  if (Tcl_PkgProvide(interp, ab1("Mmdispcmds"), ab2(MMDISPCMDS_VERSION)) 
      != TCL_OK) {
    RETURN_TCL_ERROR;
  }
  if (Oc_InitScript(interp, "Mmdispcmds", MMDISPCMDS_VERSION) != TCL_OK) {
    RETURN_TCL_ERROR;
  }
  return TCL_OK;

#undef RETURN_TCL_ERROR

}
