# FILE: opendlg.tcl
#
# Open file dialog box for mmDisp
#
# Callbacks: OPEN
Oc_Class OpenDlg {
    const public variable winpath
    const public variable outer_frame_options = "-bd 2 -relief raised"
    const public variable filter = "*.o?f *.svf *.vio"
    const public variable dialog_title = "Open File"

    const public variable menulist = {} ;# This should be a two
    ## item list: menu name + item name.
    const public variable menuraise = 1
    const public variable menucolor = red
    const private variable menudeletecleanup = Oc_Nop

    public variable callback = Oc_Nop
    private variable filewidget = {}
    private variable autobox = {}
    private variable ctrlwidget = {}
    private variable selectionwidget = {}
    private variable dirwidget = {}

    # The following variables control filename checks performed on
    # SELECT action before calling $callback.
    public variable dir_must_exist = 1
    public variable file_must_exist = 1
    public variable valid_file_types = {file fifo socket}
    ## Set valid_file_types to {} to disable file type checking.
    ## Refer to the Tcl 'file type' command for a list of possible
    ## file types.  NOTE: If 'file type' returns "link", and "link"
    ## is not on this list, then Oc_ResolveLink is called to
    ## the underlying file type, which is then recompared to this
    ## list.  If you don't *ever* want links, then put "link" on
    ## this list and check the file type explicitly in the callback.

    Constructor {args} {
	eval $this Configure -winpath .w$this $args
	# Create toplevel frame to hold dialog
	toplevel $winpath
	wm group $winpath .
	wm title $winpath $dialog_title
	if {![string match {} [info commands Ow_PositionChild]]} {
	    Ow_PositionChild $winpath  ;# Position at +.25+.25 over '.'
	}
	wm protocol $winpath WM_DELETE_WINDOW \
		"$this SwitchBoard $this CLOSE"

	bind $winpath <Destroy> "+$this WinDestroy %W"

	# Bind Ctrl-. shortcut to raise and transfer focus to '.'
	bind $winpath <Control-Key-period> {Ow_RaiseWindow .}

	focus $winpath

	# Temporarily disable callbacks until construction is complete
	set callback_keep $callback ;	set callback Oc_Nop

	# Create subwidgets
	Ow_FileSelect New filewidget $winpath \
		-callback "$this SwitchBoard" \
		-filter $filter
	set ctrlframe [frame $winpath.ctrlframe]
	AutoBox New autobox $winpath
	## Note: The AutoBox class is defined below
	CtrlWBrowse New ctrlwidget $winpath -callback "$this SwitchBoard"
	Ow_SelectBox New selectionwidget $winpath \
		-callback "$this SwitchBoard"
	Ow_DirSelect New dirwidget $winpath -callback "$this SwitchBoard"

	# Initialize selection path
	$selectionwidget SetPath [$dirwidget Cget -value]

	# Pack subwidgets
	pack [$filewidget Cget -winpath] -side left \
		-fill both -expand 1 -padx 2m -pady 2m
	pack [$ctrlwidget Cget -winpath] -in $ctrlframe \
		-side right -expand 0 -anchor ne -padx 2m -pady 2m
	pack [$autobox Cget -winpath] -in $ctrlframe \
		-side top -expand 0 -padx 2m -pady 2m
	pack $ctrlframe -side top \
		-fill x -expand 0 -anchor ne -padx 2m -pady 2m
	pack [$selectionwidget Cget -winpath] -side top \
		-fill x -expand 0 -anchor n -padx 2m -pady 2m
	pack [$dirwidget Cget -winpath] -side right \
		-fill both -expand 1 -padx 2m -pady 2m

	# Setup handling of associated menu item
	if {[llength $menulist]==2} {
	    set name [lindex $menulist 0]
	    set item [lindex $menulist 1]
	    if {$menuraise} {
		set origcmd [$name entrycget $item -command]
		set menuorigfgcolor [$name entrycget $item -foreground]
		set menuorigafgcolor [$name entrycget $item -activeforeground]
		set menudeletecleanup \
			[list $name entryconfigure $item \
			  -command $origcmd \
			  -foreground $menuorigfgcolor\
			  -activeforeground $menuorigafgcolor]
		$name entryconfigure $item \
			-command "Ow_RaiseWindow $winpath" \
			-foreground $menucolor \
			-activeforeground $menucolor
	    } else {
		$name entryconfigure $item -state disabled
		set menudeletecleanup \
			[list $name entryconfigure $item -state normal]
	    }
	}

	# Re-enable callbacks
	set callback $callback_keep
    }
    method GetOptionState {} { autobox } {
	# Returns a list of autobox state pairs,
	# e.g., -autosample 0 -autozoom 1
	return [$autobox GetState]
    }
    method GetFilename {} { selectionwidget } {
	# Returns full path+filename of selected value
	return [$selectionwidget ReadDisplayedFullname]
    }
    method SwitchBoard { subwidget actionid args } {
	# General callback routing routine

	switch -exact -- $actionid {
	    UPDATE {
		# Update actions
		if {[string compare $subwidget $filewidget]==0} {
		    $selectionwidget SetFilename [lindex $args 1]
		} elseif {[string compare $subwidget $dirwidget]==0} {
		    $filewidget SetPath [lindex $args 1]
		    $selectionwidget SetPath [lindex $args 1]
		} else {
		    error "Unknown UPDATE callback source: $subwidget"
		}
	    }
	    SELECT {
		# Check that selected filename is valid
		set fname [$this GetFilename]
		if {$dir_must_exist \
			&& ![file isdirectory [file dirname $fname]]} {
		    error "Filename error: [file dirname $fname]\
			    is not a valid directory"
		}
		if {[file exists $fname]} {
		    # File exists.  Check that it is of valid type
		    if {![string match {} $valid_file_types]} {
			set ftype [file type $fname]
			if {[lsearch -exact $valid_file_types $ftype]<0} {
			    set badtype 1
			    if {[string match "link" $fname]} {
				# Resolve link and compare that type
				if {![catch {Oc_ResolveLink $fname} ttype] \
					&& [lsearch -exact \
					    $valid_file_types $ttype]<0} {
				    set ftype $ttype
				    set badtype 0
				}
			    }
			    if {$badtype} {
				error "File $fname has invalid type:\
					$ftype\nShould be one of:\
					$valid_file_types"
			    }

			}
		    }
		} elseif {$file_must_exist} {
		    error "File $fname does not exist"
		}

		set errmsg [eval $callback $this OPEN]
		if { ![$ctrlwidget Cget -browse_state] \
			&& [string match {} $errmsg] } {
		    $this Delete ; return
		}
	    }
	    CLOSE {
		$this Delete ; return
	    }
	    RESCAN {
		$filewidget Rescan
		$dirwidget Rescan
	    }
	    DELETE {}
	    default { error "Unknown actionid: $actionid" }
	}
    }
    method WinDestroy { w } { winpath } {
	if {[info exists winpath] && [string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this Delete
    }
    private variable delete_in_progress = 0
    Destructor {
	# Protect against re-entrancy
	if { [info exists delete_in_progress] && \
		$delete_in_progress } { return }
	set delete_in_progress 1

	# Notify client of impending destruction
	catch {eval $callback $this DELETE}

	# Reset associated menu item
        catch { eval $menudeletecleanup }

	# Delete subwidgets
	catch {$filewidget      Delete}
	catch {$autobox         Delete}
	catch {$ctrlwidget      Delete}
	catch {$selectionwidget Delete}
	catch {$dirwidget       Delete}

	# Return focus to main window
	Ow_MoveFocus .

	# Delete window
	if {[info exists winpath] && [winfo exists $winpath]} {
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}

# Open parameter "auto" options box for mmDisp File|Open dialog
#
# NOTE: This subwidget does not make callbacks.
;Oc_Class AutoBox {
    const public variable winpath
    const public variable outer_frame_options = "-bd 4 -relief ridge"
    public variable autosample = 1
    public variable autodatascale = 1
    public variable autozoom = 1
    public variable autoslice = 1
    Constructor {basewinpath args} {
	if {![string match {.} \
		[string index $basewinpath \
		[expr {[string length $basewinpath]-1}]]]} {
	    append basewinpath {.}
	}
	eval $this Configure -winpath ${basewinpath}w$this $args

	# Pack entire subwidget into a subframe.
	eval frame $winpath $outer_frame_options
	set innerframe [frame $winpath.innerframe -relief raised -bd 2]
	set checkframe [frame $winpath.checkframe -relief sunken -bd 2]

	set autolabel [label $winpath.label -text "Auto:" \
		-relief flat -padx 4 -pady 4]

	set but1 [checkbutton $winpath.btn1 -text "Subsample" \
		-relief flat -anchor w \
		-variable [$this GlobalName autosample]]
	set but2 [checkbutton $winpath.btn2 -text "Data Scale" \
		-relief flat -anchor w \
		-variable [$this GlobalName autodatascale]]
	set but3 [checkbutton $winpath.btn3 -text "Zoom" \
		-relief flat -anchor w \
		-variable [$this GlobalName autozoom]]
	set but4 [checkbutton $winpath.btn4 -text "Slice" \
		-relief flat -anchor w \
		-variable [$this GlobalName autoslice]]

	pack $innerframe
	pack $autolabel $checkframe -in $innerframe \
		-side left -anchor ne
	pack $but1 $but2 $but3 $but4 -in $checkframe \
		-side top -expand 1 -fill both

	# Widget bindings
	bind $winpath <Destroy> "$this WinDestroy %W"
    }
    method GetState {} { autosample autodatascale autozoom autoslice } {
	set statelist {}
	foreach i { autosample autodatascale autozoom autoslice } {
	    lappend statelist -$i [set $i]
	}
	return $statelist
    }
    method WinDestroy { w } { winpath } {
	if {[info exists winpath] && [string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this Delete
    }
    Destructor {
	if {[info exists winpath] && [winfo exists $winpath]} {
            # Remove <Destroy> binding
	    bind $winpath <Destroy> {}
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}
