/* FILE: ext.cc                 -*-Mode: c++-*-
 *
 * Base OXS extension class.  This is an abstract class.
 *
 */


#include <iostream>
#include <stdlib.h>
#include "director.h"
#include "ext.h"
#include "labelvalue.h"
#include "oc.h"
#include "nb.h"
#include "util.h"

/* End includes */

Oxs_Ext::~Oxs_Ext()
{ 
#ifdef OC_STL_MAP_BROKEN_EMPTY_DELETE
  // Some STL implementations (MVC++ 6.0 on Windows/Alpha) have a bug
  // that causes a hang if init_strings is destroyed when
  // empty.  To work around this, we just insert a dummy element.
  init_strings[string("foo")] = string("bar");
#endif
}

Oxs_Ext::Oxs_Ext(const char* idname,Oxs_Director* dtr)
  : instanceId(idname), director(dtr) {}

// Helper function for extended constructor (see below)
void Oxs_Ext::AddNewInitStringPair
(const string& key,
 const string& value) 
{ // If key is already in init_strings, throw an exception.
  // Otherwise, add key+value to init_strings.
  map<string,string>::const_iterator it = init_strings.find(key);
  if(it!=init_strings.end()) {
    string msg = string("Duplicate label field \"");
    msg += key;
    msg += string("\" in Specify block.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  init_strings[key] = value;
}

// Extended constructor with built-in argument parsing
Oxs_Ext::Oxs_Ext(const char* idname,Oxs_Director* dtr,
                 const char* argstr)
  : instanceId(idname), director(dtr)
{
  // Use Tcl_SplitList to break argstr into components
  Oxs_SplitList arglist;
  if(arglist.Split(argstr)!=TCL_OK) {
    // Input string not a valid Tcl list; Throw exception
    string msg = string("Format error in Specify block---"
			"block is not a valid Tcl list: ");
    msg += string(argstr);
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  if(arglist.Count()%2 != 0) {
    // Input string is not a list of pairs; Throw exception
    char temp_buf[1024];
    Oc_Snprintf(temp_buf,sizeof(temp_buf),
		"Format error in Specify block---"
		"block should be composed of label+value pairs,"
		" but element count is %d, which is not even: ",
		arglist.Count());
    string msg=string(temp_buf) + string(argstr);
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  // Fill in init_strings key-value array
  for(int i=0;i+1<arglist.Count();i+=2) {
    string key = arglist[i];
    string value = arglist[i+1];

    // Skip "comment" key-value pairs
    if(key.compare("comment")==0) continue;

    // Expand attribute objects
    if(key.compare("attributes")==0) {
      Oxs_SplitList split;
      if(split.Split(value.c_str())!=TCL_OK) {
        // Input string not a valid Tcl list; Throw exception
        string msg = string("MIF input parameter value associated"
                            " with label \"");
	msg += key;
	msg += string("\" is not a valid Tcl list: ");
	msg += value;
        throw Oxs_Ext::Error(this,msg.c_str());
      }
      vector<string> params;
      split.FillParams(params);
      Oxs_OwnedPointer<Oxs_LabelValue> attr;
      OXS_GET_EXT_OBJECT(params,Oxs_LabelValue,attr);

      // "attr" holds label-value pairs, which we now
      // place into init_strings.
      vector<string> labels;
      attr->GetLabels(labels);
      vector<string>::const_iterator it = labels.begin();
      while(it != labels.end()) {
        AddNewInitStringPair(*it,attr->GetValue(*it));
        ++it;
      }
      // "attr" is automatically cleaned up at block exit.
    } else {
      // Store value against key
      AddNewInitStringPair(key,value);
    }
  }
}

void
Oxs_Ext::CheckInitValueParamCount(const string& key,UINT4m count) const
{ // Throws an exception if key is not in init_strings, or if the
  // associated value is not a valid Tcl list, or if the number
  // of parameters in the value is not exactly count.
  map<string,string>::const_iterator it = init_strings.find(key);
  if(it==init_strings.end()) {
    string msg =
      string("MIF input block is missing specification for label \"");
    msg += key;
    msg += string("\"");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  const string& value = it->second;
  Oxs_SplitList params;
  if(params.Split(value.c_str())!=TCL_OK) {
    // Input string not a valid Tcl list; Throw exception
    string msg = string("MIF input parameter value associated "
                        "with label \"");
    msg += key;
    msg += string("\" is not a valid Tcl list: ");
    msg += value;
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  if(static_cast<UINT4m>(params.Count())!=count) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
                "MIF input block label \"%s\" specification"
                " has %d item(s); it should have exactly %d item(s).",
                key.c_str(),params.Count(),count);
    throw Oxs_Ext::Error(this,buf);
  }
}

BOOL Oxs_Ext::HasInitValue(const string& key) const
{
  map<string,string>::const_iterator it = init_strings.find(key);
  if(it==init_strings.end()) return 0;
  return 1;
}



BOOL Oxs_Ext::FindInitValue(const string& key,string& value) const
{
  map<string,string>::const_iterator it = init_strings.find(key);
  if(it==init_strings.end()) return 0;
  value = it->second;
  return 1;
}

BOOL
Oxs_Ext::FindInitValue(const string& key,vector<string>& params) const
{
  string value;
  if(!FindInitValue(key,value)) return 0;
  Oxs_SplitList split;
  if(split.Split(value.c_str())!=TCL_OK) {
    // Input string not a valid Tcl list; Throw exception
    string msg = string("MIF input parameter value associated"
                        " with label \"");
    msg += key;
    msg += string("\" is not a valid Tcl list: ");
    msg += value;
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  split.FillParams(params);
  return 1;
}

void
Oxs_Ext::FindRequiredInitValue(const string& key,string& value) const
{
  if(!FindInitValue(key,value)) {
    string msg = string("Missing required field \"");
    msg += key;
    msg += string("\" in Specify block.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
}

void
Oxs_Ext::FindRequiredInitValue
(const string& key,
 vector<string>& params) const
{
  if(!FindInitValue(key,params)) {
    string msg = string("Missing required field \"");
    msg += key;
    msg += string("\" in Specify block.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
}

REAL8m Oxs_Ext::GetRealInitValue(const string& key)
{
  string value;
  FindRequiredInitValue(key,value);
  BOOL error;
  REAL8m result = Nb_Atof(value.c_str(),error);
  if(error) {
    string msg = string("Value associated with MIF input label \"");
    msg += key;
    msg += string("\" is not a real number: ");
    msg += value;
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  DeleteInitValue(key);
  return result;
}

REAL8m Oxs_Ext::GetRealInitValue(const string& key,
                                 REAL8m default_value)
{
  if(!HasInitValue(key)) return default_value;
  return GetRealInitValue(key);
}

ThreeVector Oxs_Ext::GetThreeVectorInitValue(const string& key)
{
  vector<string> value;
  FindRequiredInitValue(key,value);
  if(value.size()!=3) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
                "MIF input block label \"%s\" value specification"
                " has %d item(s); it should have exactly 3 item(s).",
                key.c_str(),value.size());
    throw Oxs_Ext::Error(this,buf);
  }
  ThreeVector vec;
  BOOL err1,err2,err3;
  vec.x=Nb_Atof(value[0].c_str(),err1);
  vec.y=Nb_Atof(value[1].c_str(),err2);
  vec.z=Nb_Atof(value[2].c_str(),err3);
  if(err1 || err2 || err3) {
    string msg = string("MIF input block value specification"
			" associated with label \"");
    msg += key;
    msg += string("\" is not a list of 3 real numbers: ");
    msg += value[0];
    msg += string(" ");
    msg += value[1];
    msg += string(" ");
    msg += value[2];
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  DeleteInitValue(key);
  return vec;
}

ThreeVector
Oxs_Ext::GetThreeVectorInitValue(const string& key,
                                 const ThreeVector& default_value)
{
  if(!HasInitValue(key)) return default_value;
  return GetThreeVectorInitValue(key);
}

INT4m Oxs_Ext::GetIntInitValue(const string& key)
{
  string value;
  FindRequiredInitValue(key,value);
  char* endptr;
  long int result = strtol(value.c_str(),&endptr,10);
  if(value.size()==0 || *endptr!='\0') {
    string msg = string("Value associated with MIF input label \"");
    msg += key;
    msg += string("\" is not a integer: ");
    msg += value;
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  DeleteInitValue(key);
  return static_cast<INT4m>(result);
}

INT4m Oxs_Ext::GetIntInitValue(const string& key,
                               INT4m default_value)
{
  if(!HasInitValue(key)) return default_value;
  return GetIntInitValue(key);
}

UINT4m Oxs_Ext::GetUIntInitValue(const string& key)
{
  string value;
  FindRequiredInitValue(key,value);
  char* endptr;
  unsigned long int result = strtoul(value.c_str(),&endptr,10);
  if(value.size()==0 || *endptr!='\0') {
    string msg = string("Value associated with MIF input label \"");
    msg += key;
    msg += string("\" is not an unsigned integer: ");
    msg += value;
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  DeleteInitValue(key);
  return static_cast<UINT4m>(result);
}

UINT4m Oxs_Ext::GetUIntInitValue(const string& key,
                               UINT4m default_value)
{
  if(!HasInitValue(key)) return default_value;
  return GetUIntInitValue(key);
}

string Oxs_Ext::GetStringInitValue(const string& key)
{
  string value;
  FindRequiredInitValue(key,value);
  DeleteInitValue(key);
  return value;
}

string Oxs_Ext::GetStringInitValue(const string& key,
                                   const string& default_value)
{
  if(!HasInitValue(key)) return default_value;
  return GetStringInitValue(key);
}

void
Oxs_Ext::GetRealVectorInitValue
(const string& key,
 UINT4m length,
 vector<REAL8m>& values)
{ // The value associated with key should be a list of "length"
  // real values.  This function converts the value into a vector
  // of REAL8m's, stores the results in export values, and
  // removes the key-value pair from init_strings.  Throws an
  // exception if key in not in init_strings, of if any of the
  // values cannot be completely converted to a REAL8m, or if the
  // parameter list associated with key does not have "length"
  // entries.
  vector<string> strvals;
  FindRequiredInitValue(key,strvals);
  if(strvals.size()!=length) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
                "Value for label \"%s\" in Specify block"
                " has wrong number of elements:"
                " %d (should be %d)",
                key.c_str(),strvals.size(),length);
    throw Oxs_Ext::Error(this,buf);
  }

  values.clear();
  vector<string>::const_iterator it = strvals.begin();
  while(it != strvals.end()) {
    BOOL error;
    values.push_back(Nb_Atof(it->c_str(),error));
    if(error) {
      string msg = string("Entry \"");
      msg += *it;
      msg += string("\" in value list for value \"");
      msg += key;
      msg += string("\" is not a REAL value.");
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    ++it;
  }

  DeleteInitValue(key);
}

// Group expand/noexpand escape keys
static const char* OXSEXT_EXPAND_KEY   = ":expand:";
static const char* OXSEXT_NOEXPAND_KEY = ":noexpand:";

static void ExpandList(const vector<string>& grouped_list,
                       vector<string>& expanded_list) {
  // The grouped_list import to ExpandList must be a proper
  // grouped list, i.e., it must have at least 2 elements,
  // and the last element must be a positive integer.
  //  ExpandList will examine each element to see if it is
  // also a proper grouped list, and if so make a recursive
  // call to expand out the element.  An element is identified
  // as grouped list if it is a list with at least 2 elements,
  // and the last element is an integer.  For escape purposes,
  // if the last element of a sublist is the string ":noexpand:",
  // then that element is removed and the rest of the sublist
  // is used without expansion.

  if(&grouped_list == &expanded_list) {
    OXS_THROW(Oxs_BadParameter,
              string("Inputs to ExpandList must be "
                     "distinct vector<string>'s"));
  }

  expanded_list.clear();
  // Check size of group_list
  vector<string>::size_type size = grouped_list.size();
  if(size<2) {
    string msg = string("Invalid grouped list in ExpandList(),"
                        " fewer than 2 elements: ");
    if(size<1) msg += string("{<empty list>}");
    else       msg += grouped_list[0];
    OXS_THROW(Oxs_BadParameter,msg);
  }

  // Extract and check group count
  const char* countstr=grouped_list[size-1].c_str();
  char* endptr;
  long int group_count = strtol(countstr,&endptr,10);
  if(countstr[0]=='\0' || *endptr!='\0' || group_count<1) {
    string msg = string("Invalid group count in ExpandList(): ");
    msg += Oxs_MergeList(&grouped_list);
    OXS_THROW(Oxs_BadParameter,msg);
  }

  // Fill expanded_list with one copy of grouped_list
  Oxs_SplitList check_list;
  vector<string>::size_type index;
  vector<string> subgroup;
  vector<string> tmplist;
  for(index=0; index<size-1; index++) {
    check_list.Split(grouped_list[index].c_str());
    int check_list_size = check_list.Count();
    if(check_list_size<2) {
      // Single item
      expanded_list.push_back(grouped_list[index]);
    } else {
      const char* lastitem = check_list[check_list_size-1];
      if(strcmp(lastitem,OXSEXT_NOEXPAND_KEY)==0) {
	// Escape out of recursion.  Throw out trailing
	// OXSEXT_NOEXPAND_KEY element.
	for(int subindex=0;subindex<check_list_size-1;subindex++) {
	  expanded_list.push_back(check_list[subindex]);
	}
      } else {
	// See if last item is a positive integer
	char* subendptr;
	long int subgroup_count = strtol(lastitem,&subendptr,10);
	if(lastitem[0]=='\0' || *subendptr!='\0' || subgroup_count<1) {
	  // Last item is not a positive integer, so treat
	  // whole sublist as a single element.
	  expanded_list.push_back(grouped_list[index]);
        } else {
	  // Sub-group. Make recursive call
	  check_list.FillParams(subgroup);
	  ExpandList(subgroup,tmplist); // <<< RECURSIVE CALL >>>
	  expanded_list.insert(expanded_list.end(),
			       tmplist.begin(),tmplist.end());
	}
      }
    }
  }

  // Make requested number (group_count) of copies
  vector<string>::size_type base_size=expanded_list.size();
  for(long int copy_number=1; copy_number<group_count; copy_number++) {
    expanded_list.insert(expanded_list.end(),
                         expanded_list.begin(),
                         expanded_list.begin()+base_size);
  }
  
}

void
Oxs_Ext::GetGroupedStringListInitValue
(const string& key,
 vector<string>& expanded_list)
{
  // The value associated with key should be a grouped list of strings.
  // This function loads the strings into the export vector<string>
  // expanded_list, and removes the key-value pair from init_strings.
  // An exception is thrown if key in not in init_strings, if the
  // associated value does not contain at least one element, of if there
  // is a structural problem with the grouped list.  If the associated
  // value is a list of at least 2 elements, and if the last element is
  // ":expand:", then that last element is pitched and the remainder is
  // expanded as a grouped sublist.  Grouped sublists have the form
  // {x1 x2 ... xn R} where x1 through xn are strings, and R is an
  // integer repeat count.  The sublists are expanded recursively, so,
  // i.e., xi can also be either an individual value or a grouped
  // sublist of the same form.  As an example, the value "fred {barney
  // {wilma betty 3} 2} :expand:" would be expanded into "fred barney
  // wilma betty wilma betty wilma betty barney wilma betty wilma betty
  // wilma betty".  Note that there is no group count expected or
  // allowed at the top level, i.e., to get the expanded list "fred
  // fred fred" an extra level of grouping is necessary, i.e.,
  // "{fred 3} :expand:" is correct, not "fred 3 :expand:".
  //   If a sublist element has embedded spaces, and the last bit can be
  // interpreted as an integer, then the algorithm outlined above will
  // cause the sublist to be improperly expanded.  For example, consider
  // the flat list "{fred 3} {fred 3}".  This is *not* equivalent to the
  // grouped list "{{fred 3} 2} :expand:", because as a result of
  // recursive expansion, the latter expands into "fred fred fred fred
  // fred fred".  To disable recursive expansion use the ":noexpand:"
  // keyword in the repeat count location.  The proper grouped list
  // representation of "{fred 3} {fred 3}" is
  //             "{{{fred 3} :noexpand:} 2} :expand:"
  // The ":noexpand:" keyword may also be used at the top-level to force
  // a flat, i.e., unexpanded list.  As explained above, no-expansion is
  // the default, but ":noexpand:" can be used to handle the situation
  // where a flat list by chance has ":expand:" as the last element.  The
  // ":noexpand:" keyword may be safely added to any flat list, which also
  // protects against accidental eating of a chance ":noexpand:" last
  // element in a flat list.  This is mainly useful in lists generated
  // by code from unknown input.

  vector<string> strvals;
  FindRequiredInitValue(key,strvals);
  if(strvals.size()<1) {
    string msg = string("Value for label \"");
    msg += key;
    msg += string("\" in Specify block is an empty list.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  // Is this a grouped list?
  if(strvals.size()<2 || strvals.back().compare(OXSEXT_EXPAND_KEY)!=0) {
    // Flat list.  If last element is OXSEXT_NOEXPAND_KEY, remove it.
    if(strvals.back().compare(OXSEXT_NOEXPAND_KEY)==0) {
      strvals.pop_back();
    }
    swap(expanded_list,strvals);
  } else {
    // Grouped list.  Replace OXSEXT_EXPAND_KEY keyword with the
    // implicit full string group count of 1.
    strvals.back() = string("1");
    
    // Expand
    try {
      ExpandList(strvals,expanded_list);
    } catch (Oc_Exception& errmsg) {
      string msg = string("Error expanding group structure for label \"");
      msg += key;
      msg += string("\": ");
      msg += string(errmsg.MessageType());
      msg += string(":\n");
      msg += errmsg.MessageText();
      throw Oxs_Ext::Error(this,msg.c_str());
    }
  }

  DeleteInitValue(key);
}

void
Oxs_Ext::GetGroupedRealListInitValue
(const string& key,
 vector<REAL8m>& export_list)
{ // Analogous to GetGroupedStringListInitValue(), but the final
  // expanded list should consist entirely of REAL8m values.  An
  // exception is thrown if key in not in init_strings, if the
  // associated value does not contain at least one element, if there is
  // a structural problem with the grouped list, of if any of the values
  // in the expanded_list cannot be completely converted to a REAL8m
  // value. See notes in the GetGroupedStringListInitValue() function
  // for additional details.
  vector<string> strvals;
  FindRequiredInitValue(key,strvals);
  if(strvals.size()<1) {
    string msg = string("Value for label \"");
    msg += key;
    msg += string("\" in Specify block is an empty list.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  vector<string> expanded_list;

  // Is this a grouped list?
  if(strvals.size()<2 || strvals.back().compare(OXSEXT_EXPAND_KEY)!=0) {
    // Flat list.  If last element is OXSEXT_NOEXPAND_KEY, remove it.
    if(strvals.back().compare(OXSEXT_NOEXPAND_KEY)==0) {
      strvals.pop_back();
    }
    swap(expanded_list,strvals);
  } else {
    // Grouped list.  Replace OXSEXT_EXPAND_KEY keyword with the
    // implicit full string group count of 1.
    strvals.back() = string("1");
    try {
      ExpandList(strvals,expanded_list);
    } catch (Oc_Exception& errmsg) {
      string msg = string("Error expanding group structure for label \"");
      msg += key;
      msg += string("\": ");
      msg += string(errmsg.MessageType());
      msg += string(":\n");
      msg += errmsg.MessageText();
      throw Oxs_Ext::Error(this,msg.c_str());
    }
  }

  export_list.clear();
  vector<string>::const_iterator it = expanded_list.begin();
  while(it != expanded_list.end()) {
    BOOL error;
    export_list.push_back(Nb_Atof(it->c_str(),error));
    if(error) {
      string msg = string("Entry \"");
      msg += *it;
      msg += string("\" in value list for label \"");
      msg += key;
      msg += string("\" is not a REAL value.");
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    ++it;
  }

  DeleteInitValue(key);
}

void
Oxs_Ext::GetGroupedIntListInitValue
(const string& key,
 vector<INT4m>& export_list)
{ // Analogous to GetGroupedStringListInitValue(), but the final
  // expanded list should consist entirely of INT4m values.  An
  // exception is thrown if key in not in init_strings, if the
  // associated value does not contain at least one element, if there is
  // a structural problem with the grouped list, of if any of the values
  // in the expanded_list cannot be completely converted to an INT4m
  // value. See notes in the GetGroupedStringListInitValue() function
  // for additional details.
  vector<string> strvals;
  FindRequiredInitValue(key,strvals);
  if(strvals.size()<1) {
    string msg = string("Value for label \"");
    msg += key;
    msg += string("\" in Specify block is an empty list.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  vector<string> expanded_list;

  // Is this a grouped list?
  if(strvals.size()<2 || strvals.back().compare(OXSEXT_EXPAND_KEY)!=0) {
    // Flat list.  If last element is OXSEXT_NOEXPAND_KEY, remove it.
    if(strvals.back().compare(OXSEXT_NOEXPAND_KEY)==0) {
      strvals.pop_back();
    }
    swap(expanded_list,strvals);
  } else {
    // Grouped list.  Replace OXSEXT_EXPAND_KEY keyword with the
    // implicit full string group count of 1.
    strvals.back() = string("1");
    try {
      ExpandList(strvals,expanded_list);
    } catch (Oc_Exception& errmsg) {
      string msg = string("Error expanding group structure for label \"");
      msg += key;
      msg += string("\": ");
      msg += string(errmsg.MessageType());
      msg += string(":\n");
      msg += errmsg.MessageText();
      throw Oxs_Ext::Error(this,msg.c_str());
    }
  }

  export_list.clear();
  vector<string>::const_iterator it = expanded_list.begin();
  while(it != expanded_list.end()) {
    char* endptr;
    const char* cptr=it->c_str();
    long int iresult = strtol(cptr,&endptr,10);
    if(cptr[0]=='\0' || *endptr!='\0') {
      // Error!
      string msg = string("Entry \"");
      msg += *it;
      msg += string("\" in value list for label \"");
      msg += key;
      msg += string("\" is not an INT value.");
      throw Oxs_Ext::Error(this,msg.c_str());
    } else {
      export_list.push_back(iresult);
    }
    ++it;
  }

  DeleteInitValue(key);
}

void
Oxs_Ext::GetGroupedUIntListInitValue
(const string& key,
 vector<UINT4m>& export_list)
{ // Analogous to GetGroupedStringListInitValue(), but the final
  // expanded list should consist entirely of UINT4m values.  An
  // exception is thrown if key in not in init_strings, if the
  // associated value does not contain at least one element, if there is
  // a structural problem with the grouped list, of if any of the values
  // in the expanded_list cannot be completely converted to an UINT4m
  // value. See notes in the GetGroupedStringListInitValue() function
  // for additional details.
  vector<string> strvals;
  FindRequiredInitValue(key,strvals);
  if(strvals.size()<1) {
    string msg = string("Value for label \"");
    msg += key;
    msg += string("\" in Specify block is an empty list.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  vector<string> expanded_list;

  // Is this a grouped list?
  if(strvals.size()<2 || strvals.back().compare(OXSEXT_EXPAND_KEY)!=0) {
    // Flat list.  If last element is OXSEXT_NOEXPAND_KEY, remove it.
    if(strvals.back().compare(OXSEXT_NOEXPAND_KEY)==0) {
      strvals.pop_back();
    }
    swap(expanded_list,strvals);
  } else {
    // Grouped list.  Replace OXSEXT_EXPAND_KEY keyword with the
    // implicit full string group count of 1.
    strvals.back() = string("1");
    try {
      ExpandList(strvals,expanded_list);
    } catch (Oc_Exception& errmsg) {
      string msg = string("Error expanding group structure for label \"");
      msg += key;
      msg += string("\": ");
      msg += string(errmsg.MessageType());
      msg += string(":\n");
      msg += errmsg.MessageText();
      throw Oxs_Ext::Error(this,msg.c_str());
    }
  }

  export_list.clear();
  vector<string>::const_iterator it = expanded_list.begin();
  while(it != expanded_list.end()) {
    char* endptr;
    const char* cptr=it->c_str();
    unsigned long int uiresult = strtoul(cptr,&endptr,10);
    if(cptr[0]=='\0' || *endptr!='\0') {
      // Error!
      string msg = string("Entry \"");
      msg += *it;
      msg += string("\" in value list for label \"");
      msg += key;
      msg += string("\" is not an UINT value.");
      throw Oxs_Ext::Error(this,msg.c_str());
    } else {
      export_list.push_back(uiresult);
    }
    ++it;
  }

  DeleteInitValue(key);
}

BOOL
Oxs_Ext::DeleteInitValue(const string& key)
{ // Removes the specified key-value pair from init_strings.
  //  Returns 1 on success, 0 if key was not found in init_strings.
  // It seems that 'init_strings.erase(key)' should work here, but
  //  this produces a NULL-pointer de-reference under MVC++ 6.0 on
  //  Windows/Alpha.  So we just use find() and erase(it) instead.
  map<string,string>::iterator it = init_strings.find(key);
  if(it==init_strings.end()) return 0;
  init_strings.erase(it);
  return 1;
}

void
Oxs_Ext::FindOrCreateObject
(const vector<string>& params,
 Oxs_OwnedPointer<Oxs_Ext>& obj) const
{
  if(params.size()<1 || params.size()>2) {
    // Note: The following construction of msg is broken up
    // into 3 separate statements, because when merged together
    // as a single statement with multiple +'s the Intel C++ 5
    // compiler deletes the string temporary associated with
    // InstanceName() too soon.
    string msg = string("Invalid parameter string detected in"
			" Oxs_Ext reference-or-create sub-block"
			" in object ");
    msg += string(InstanceName());
    msg += string("; Wrong number of parameters (should be 1 or 2)");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  Oxs_Ext* objptr=NULL;
  if(params.size()==1) {
    objptr = director->FindExtObject(params[0]);
    if(objptr==NULL) {
      // See note above about Intel C++ 5 compiler
      string msg = "In Specify block for ";
      msg += string(InstanceName());
      msg += string(": Unable to find pre-existing Oxs_Ext object ");
      msg += params[0];
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    obj.SetAsNonOwner(objptr);
  } else {
    try {
      objptr = MakeNew(params[0].c_str(),director,params[1].c_str());
    } catch (Oxs_Ext::Error& err) {
      // See note above about Intel C++ 5 compiler
      string msg = "In Specify block for ";
      msg += string(InstanceName());
      msg += string(": Unable to create inline Oxs_Ext object ");
      msg += params[0];
      msg += string(" with args ");
      msg += params[1];
      msg += string(" --- \n");
      msg += string(err);
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    obj.SetAsOwner(objptr);
  }
}

void Oxs_Ext::VerifyAllInitArgsUsed()
{ // Throws an exception if GetInitValueCount() != 0.  Use this in
  // combination with DeleteInitValue() to check MIF input block for
  // superfluous parameters.  This should be called only the actual
  // instantiating class (not a parenting class).
  if(GetInitValueCount()!=0) {
    string msg =  string("Specify block includes the following"
                         " unrecognized or unused parameters:\n");
    map<string,string>::const_iterator it = init_strings.begin();
    while(it!=init_strings.end()) {
      vector<string> paramline;
      paramline.push_back(it->first);
      paramline.push_back(it->second);
      msg += "   " + Oxs_MergeList(&paramline) + "\n";
      ++it;
    }
    throw Oxs_Ext::Error(this,msg.c_str());
  }
}

string Oxs_Ext::ExtractClassName(const string& name)
{
  string classname=name;
  string::size_type endpt=classname.find(':');
  if(endpt!=string::npos) {
    classname.resize(endpt); // colon found; truncate
  }
  return classname;
}

string Oxs_Ext::ExtractInstanceTail(const string& name)
{
  string::size_type index = name.find(':');
  string tail; // Default instance tail is empty string
  if(index!=string::npos) {
    tail = name.c_str()+index+1; // Copy tail
  }
  return tail;
}

BOOL Oxs_Ext::ValidateIdStrings() const
{
  string classrequest=ExtractClassName(instanceId);
  if(classrequest.compare(ClassName())!=0) {
    // Error, wrong type!
    string msg = string("Incorrect object type request (");
    msg += classrequest;
    msg += string(") in ");
    msg += string(ClassName());
    msg += string(" constructor.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  return 1;
}


// Oxs_Ext Registration code /////////////////////////////////////

Oxs_ExtRegistrationBlock::Oxs_ExtRegistrationBlock
(const char* name,Oxs_ExtChildConstructFunc* ctr)
  : classid(name), construct(ctr)
{
  Oxs_Ext::Register(*this);
}

// Function that returns a reference to an STL "map" of all registered
// classes.  The map is a static variable of RegClasses(). This insures
// the map is initialized before it is used, which would otherwise be
// difficult, because Oxs_Ext child classes call Oxs_Ext::Register
// during static initialization.  (See Section 9.4.1 of "The C++
// Programming Lanuage," Bjarne Stroustrup, 3rd edition for a discussion
// of nonlocal variable initialization order.)
map<string,Oxs_ExtRegistrationBlock>& Oxs_Ext::RegClasses()
{
  static map<string,Oxs_ExtRegistrationBlock> regclasses;
  return regclasses;
}

void Oxs_Ext::Register(Oxs_ExtRegistrationBlock const& regblk)
{
  map<string,Oxs_ExtRegistrationBlock>& regclasses=RegClasses();
  regclasses[regblk.classid]=regblk;
}

Oxs_Ext* Oxs_Ext::MakeNew
(const char* idstring, // Child object name
 Oxs_Director* dtr,    // App director
 const char* argstr)   // MIF input block params
{
  map<string,Oxs_ExtRegistrationBlock>& regclasses=RegClasses();
  string classname(ExtractClassName(idstring));
  map<string,Oxs_ExtRegistrationBlock>::iterator it =
    regclasses.find(classname);
  if(it==regclasses.end()) {
    string msg = string("In Oxs_Ext::MakeNew();"
			" unrecognized object class: ");
    msg += classname;
    throw UnregisteredType(msg.c_str());
  }
  Oxs_Ext* obj=NULL;
  try {
    obj=it->second.construct(idstring,dtr,argstr);
    obj->ValidateIdStrings();  // Verify we got the right one! (Safety)
  }
  catch (Oxs_Ext::Error& err) {
    throw ConstructFailure(err.c_str());
  }
  return obj;
}

string Oxs_Ext::ListRegisteredChildClasses()
{
  map<string,Oxs_ExtRegistrationBlock>& regclasses=RegClasses();
  string result("Oxs Registration List ---");
  map<string,Oxs_ExtRegistrationBlock>::iterator it;
  for(it=regclasses.begin();it!=regclasses.end();++it) {
    result.append("\n ");
    result.append(it->first);
  }
  return result;
}


void
Oxs_Ext::StageRequestCount(unsigned int& min,unsigned int& max) const
{
  min=0;  max=UINT_MAX; // No restriction.
}
