/* FILE: mesh.cc                 -*-Mode: c++-*-
 *
 * Abstract Oxs_Mesh class, derives from Oxs_Ext and Oxs_Lock classes.
 *
 */

#include "mesh.h"
#include "meshvalue.h"
#include "vf.h"

/* End includes */


////////////////////////////////////////////////////////////////////////
// Oxs_Mesh (generic mesh) class

// Constructors
Oxs_Mesh::Oxs_Mesh
( const char* name,     // Child instance id
  Oxs_Director* newdtr  // App director
  ) : Oxs_Ext(name,newdtr)
{}

Oxs_Mesh::Oxs_Mesh
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr    // MIF block argument string
  ) : Oxs_Ext(name,newdtr,argstr)
{}


//////////////////////////////////////////////////
// Default file output using generic Oxs_Mesh interface.
// Throws exception on error
void Oxs_Mesh::WriteOvfIrregular
(Tcl_Channel channel,   // Output channel
 const char* title,     // Long filename or title
 const char* desc,      // Description to embed in output file
 const char* valueunit, // Field units, such as "A/m".
 const char* datatype,  // Either "binary" or "text"
 const char* precision, // For binary, "4" or "8";
                        ///  for text, a printf-style format
 const Oxs_MeshValue<ThreeVector>* vec,  // Vector array
 const Oxs_MeshValue<REAL8m>* scale,     // Optional scaling for vec
                    /// Set scale to NULL to use vec values directly.
 const ThreeVector* stephints  // If non-null, then the values are
                              /// written as stephints to the file.
 ) const
{
  // Check import validity
  enum DataType { BINARY, TEXT };
  DataType dt = BINARY;
  if(strcmp("text",datatype)==0) {
    dt = TEXT;
  } else if(strcmp("binary",datatype)!=0){
    string errmsg = string("Bad datatype: \"") + string(datatype)
      + string("\"  Should be either \"binary\" or \"text\"");
    throw Oxs_Ext::Error(this,errmsg.c_str());
  }

  int datawidth = 0;
  string dataformat;
  if(dt == BINARY) {
    datawidth = atoi(precision);
    if(datawidth != 4 && datawidth != 8) {
      string errmsg = string("Bad precision: ") + string(precision)
	+ string("  Should be either 4 or 8.");
      throw Oxs_Ext::Error(this,errmsg.c_str());
    }
  } else {
    if(precision==NULL || precision[0]=='\0')
      precision="%# .17g";  // Default format
    string temp = string(precision) + string(" ");
    dataformat = temp + temp + temp + temp + temp +
      string(precision) + string("\n");
  }

  if(!vec->CheckMesh(this)) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),"Size mismatch; input data length=%u,"
		" which is different than mesh size=%u",
		vec->Size(),Size());
    throw Oxs_Ext::Error(this,buf);
  }

  if(scale!=NULL && !scale->CheckMesh(this)) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),"Size mismatch; input scale data length=%u,"
		" which is different than mesh size=%u",
		scale->Size(),Size());
    throw Oxs_Ext::Error(this,buf);
  }

  // Write header
  Nb_FprintfChannel(channel,NULL,1024,"# OOMMF: irregular mesh v1.0\n");
  Nb_FprintfChannel(channel,NULL,1024,"# Segment count: 1\n");
  Nb_FprintfChannel(channel,NULL,1024,"# Begin: Segment\n");
  Nb_FprintfChannel(channel,NULL,1024,"# Begin: Header\n");
  if(title==NULL || title[0]=='\0')
    Nb_FprintfChannel(channel,NULL,1024,"# Title: unknown\n");
  else
    Nb_FprintfChannel(channel,NULL,1024,"# Title: %s\n",title);
  if(desc!=NULL && desc[0]!='\0') {
    // Print out description, breaking at newlines as necessary.
    // Note: This block is optional
    const char *cptr1,*cptr2;
    cptr1=desc;
    while((cptr2=strchr(cptr1,'\n'))!=NULL) {
      Nb_FprintfChannel(channel,NULL,1024,
			"# Desc: %.*s\n",(int)(cptr2-cptr1),cptr1);
      cptr1=cptr2+1;
    }
    if(*cptr1!='\0')
      Nb_FprintfChannel(channel,NULL,1024,"# Desc: %s\n",cptr1);
  }
  Nb_FprintfChannel(channel,NULL,1024,"# meshunit: m\n");
  Nb_FprintfChannel(channel,NULL,1024,"# valueunit: %s\n",valueunit);
  Nb_FprintfChannel(channel,NULL,1024,"# valuemultiplier: 1.0\n");

  Oxs_Box bbox;
  GetBoundingBox(bbox);
  Nb_FprintfChannel(channel,NULL,1024,
	  "# xmin: %.16g\n# ymin: %.16g\n# zmin: %.16g\n"
	  "# xmax: %.16g\n# ymax: %.16g\n# zmax: %.16g\n",
	  bbox.GetMinX(),bbox.GetMinY(),bbox.GetMinZ(),
	  bbox.GetMaxX(),bbox.GetMaxY(),bbox.GetMaxZ());

  // As of 6/2001, mmDisp supports display of out-of-plane rotations;
  // Representing the boundary under these conditions is awkward with
  // a single polygon.  So don't write the boundary line, and rely
  // on defaults based on the data range and step size.

  REAL8m minmag=0,maxmag=0;
  if(Size()>0) {
    minmag=DBL_MAX;
    for(UINT4m i=0;i<Size();i++) {
      REAL8m val=(*vec)[i].MagSq();
      if(scale!=NULL) {
        REAL8m tempscale=(*scale)[i];
        val*=tempscale*tempscale;
      }
      if(val<minmag && val>0) minmag=val; // minmag is smallest non-zero
      if(val>maxmag) maxmag=val;         /// magnitude.
    }
    if(minmag>maxmag) minmag=maxmag;
    maxmag=sqrt(maxmag);
    minmag=sqrt(minmag);
    minmag*=0.999;  // Underestimate lower bound by 0.1% to protect against
    /// rounding errors.
  }

  Nb_FprintfChannel(channel,NULL,1024,
		    "# ValueRangeMaxMag: %.16g\n",maxmag);
  Nb_FprintfChannel(channel,NULL,1024,
		    "# ValueRangeMinMag: %.16g\n",minmag);

  Nb_FprintfChannel(channel,NULL,1024,
		    "# meshtype: irregular\n");
  Nb_FprintfChannel(channel,NULL,1024,
		    "# pointcount: %lu\n",
          static_cast<unsigned long>(Size()));

  // Write stephints, if requested
  if(stephints!=NULL) {
    Nb_FprintfChannel(channel,NULL,1024,
		      "# xstepsize: %.16g\n",stephints->x);
    Nb_FprintfChannel(channel,NULL,1024,
		      "# ystepsize: %.16g\n",stephints->y);
    Nb_FprintfChannel(channel,NULL,1024,
		      "# zstepsize: %.16g\n",stephints->z);
  }


  Nb_FprintfChannel(channel,NULL,1024,"# End: Header\n");

  // Write data block
  if(dt == BINARY) {
    if(datawidth==4) {
      Nb_FprintfChannel(channel,NULL,1024,"# Begin: Data Binary 4\n");
      REAL4 buf[6];
      buf[0]=1234567.;  // 4-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
	ThreeVector loc;   Center(i,loc);
        buf[0] = REAL4(loc.x);
	buf[1] = REAL4(loc.y);
	buf[2] = REAL4(loc.z);
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  buf[3] = REAL4(v.x);
	  buf[4] = REAL4(v.y);
	  buf[5] = REAL4(v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  buf[3] = REAL4(tempscale*v.x);
	  buf[4] = REAL4(tempscale*v.y);
	  buf[5] = REAL4(tempscale*v.z);
	}
        Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,6);
	// Vf_OvfFileFormatSpecs::WriteBinary performs
	// byte-swapping as needed.
      }
      Nb_FprintfChannel(channel,NULL,1024,"\n# End: Data Binary 4\n");
    } else {
      Nb_FprintfChannel(channel,NULL,1024,"# Begin: Data Binary 8\n");
      REAL8 buf[6];
      buf[0]=123456789012345.;  // 8-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
	ThreeVector loc;   Center(i,loc);
        buf[0] = REAL4(loc.x);
	buf[1] = REAL4(loc.y);
	buf[2] = REAL4(loc.z);
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  buf[3] = REAL4(v.x);
	  buf[4] = REAL4(v.y);
	  buf[5] = REAL4(v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  buf[3] = REAL4(tempscale*v.x);
	  buf[4] = REAL4(tempscale*v.y);
	  buf[5] = REAL4(tempscale*v.z);
	}
        Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,6);
	// Vf_OvfFileFormatSpecs::WriteBinary performs
	// byte-swapping as needed.
      }
      Nb_FprintfChannel(channel,NULL,1024,"\n# End: Data Binary 8\n");
    }
  } else {
    Nb_FprintfChannel(channel,NULL,1024,"# Begin: Data Text\n");
    INT4m size=Size();
    for(INT4m i=0 ; i<size ; ++i) {
	ThreeVector loc;   Center(i,loc);
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  Nb_FprintfChannel(channel,NULL,1024,dataformat.c_str(),
		  loc.x,loc.y,loc.z,v.x,v.y,v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  Nb_FprintfChannel(channel,NULL,1024,dataformat.c_str(),
		  loc.x,loc.y,loc.z,
		  tempscale*v.x,tempscale*v.y,tempscale*v.z);
	}
    }
    Nb_FprintfChannel(channel,NULL,1024,"# End: Data Text\n");
  }

  Nb_FprintfChannel(channel,NULL,1024,"# End: Segment\n");
}

// Version of WriteOvf that takes a filename instead of a Tcl_Channel.
// This is for backwards compatibility and convenience.
void Oxs_Mesh::WriteOvf
(const char* filename,  // Name of output file
 const char* title,     // Long filename or title
 const char* desc,      // Description to embed in output file
 const char* valueunit, // Field units, such as "A/m".
 const char* meshtype,  // Either "rectangular" or "irregular"
 const char* datatype,  // Either "binary" or "text"
 const char* precision, // For binary, "4" or "8";
                        ///  for text, a printf-style format
 const Oxs_MeshValue<ThreeVector>* vec,  // Vector array
 const Oxs_MeshValue<REAL8m>* scale      // Optional scaling for vec
 /// Set scale to NULL to use vec values directly.
 ) const
{
  if(filename==NULL) {
    // Bad input
    string msg=string("Failure in object ")
      + string(InstanceName())
      + string("; filename not specified.");
    throw Oxs_Ext::Error(msg.c_str());
  }

  int orig_errno=Tcl_GetErrno();
  Tcl_SetErrno(0);
  Tcl_Channel channel = Tcl_OpenFileChannel(NULL,
	const_cast<char*>(filename),
	const_cast<char*>("w"),0666);
  /// We'll *ASSUME* Tcl isn't really going to modify filename/mode.
  int new_errno=Tcl_GetErrno();
  Tcl_SetErrno(orig_errno);

  if(channel==NULL) {
    // File open failure
    string msg=string("Failure in object ")
      + string(InstanceName())
      + string(" to open file ") + string(filename);
    if(new_errno!=0)
      msg += string(" for writing: ") + string(Tcl_ErrnoMsg(new_errno));
    else
      msg += string(" for writing.");
    throw Oxs_Ext::Error(msg.c_str());
  }

  try {
    // Set channel options
    Tcl_SetChannelOption(NULL,channel,Oc_AutoBuf("-buffering"),
			 Oc_AutoBuf("full"));
    Tcl_SetChannelOption(NULL,channel,Oc_AutoBuf("-buffersize"),
			 Oc_AutoBuf("100000"));
    /// What's a good size???
    if(strcmp("text",datatype)!=0) {
      // Binary mode
      Tcl_SetChannelOption(NULL,channel,Oc_AutoBuf("-encoding"),
			   Oc_AutoBuf("binary"));
      Tcl_SetChannelOption(NULL,channel,Oc_AutoBuf("-translation"),
			   Oc_AutoBuf("binary"));
    }

    if(title==NULL || title[0]=='\0') {
      title=filename;
    }
    WriteOvf(channel,title,desc,valueunit,
	     meshtype,datatype,precision,vec,scale);
  } catch (...) {
    Tcl_Close(NULL,channel);
    Nb_Remove(filename); // Delete possibly corrupt or
                        /// empty file on error.
    throw;
  }
  Tcl_Close(NULL,channel);
}

////////////////////////////////////////////////////////////////////
// Default volume summing routines using generic Oxs_Mesh interface.
REAL8m
Oxs_Mesh::VolumeSum(const Oxs_MeshValue<REAL8m>& scalar) const
{
  if(!scalar.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible scalar array import to"
			 " VolumeSum(const Oxs_MeshValue<REAL8m>&)");
  }
  const UINT4m size=Size();
  REAL8m sum=0.0;
  for(UINT4m i=0;i<size;i++) {
    sum += scalar[i] * Volume(i);
  }
  return sum;
}

ThreeVector
Oxs_Mesh::VolumeSum(const Oxs_MeshValue<ThreeVector>& vec) const
{
  if(!vec.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSum(const Oxs_MeshValue<ThreeVector>&)");
  }
  const UINT4m size=Size();
  ThreeVector sum(0.,0.,0.);
  for(UINT4m i=0;i<size;i++) {
    sum +=  Volume(i) * vec[i] ;
  }
  return sum;
}

ThreeVector
Oxs_Mesh::VolumeSum(const Oxs_MeshValue<ThreeVector>& vec,
		    const Oxs_MeshValue<REAL8m>& scale) const
{
  if(!vec.CheckMesh(this) || !scale.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSum(const Oxs_MeshValue<ThreeVector>&,"
			 " const Oxs_MeshValue<REAL8m>& scale)");
  }
  const UINT4m size=Size();
  ThreeVector sum(0.,0.,0.);
  for(UINT4m i=0;i<size;i++) {
    sum +=  ( (Volume(i) * scale[i]) * vec[i] );
  }
  return sum;
}

REAL8m
Oxs_Mesh::VolumeSumXp(const Oxs_MeshValue<REAL8m>& scalar) const
{
  if(!scalar.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible scalar array import to"
			 " VolumeSumXp(const Oxs_MeshValue<REAL8m>&)");
  }

  Nb_Xpfloat sum=0.;
  const UINT4m size=Size();
  for(UINT4m i=0;i<size;i++) {
    sum += scalar[i] * Volume(i);
  }

  return sum.GetValue();
}

ThreeVector
Oxs_Mesh::VolumeSumXp(const Oxs_MeshValue<ThreeVector>& vec) const
{
  if(!vec.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSumXp(const Oxs_MeshValue<ThreeVector>&)");
  }

  Nb_Xpfloat sum_x=0.,sum_y=0.,sum_z=0.;
  const UINT4m size=Size();
  for(UINT4m i=0;i<size;i++) {
    REAL8m vol = Volume(i);
    sum_x += vol*vec[i].x;
    sum_y += vol*vec[i].y;
    sum_z += vol*vec[i].z;
  }

  return ThreeVector(sum_x.GetValue(),sum_y.GetValue(),sum_z.GetValue());
}

ThreeVector
Oxs_Mesh::VolumeSumXp(const Oxs_MeshValue<ThreeVector>& vec,
		      const Oxs_MeshValue<REAL8m>& scale) const
{
  if(!vec.CheckMesh(this) || !scale.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSumXp(const Oxs_MeshValue<ThreeVector>&,"
			 " const Oxs_MeshValue<REAL8m>& scale)");
  }

  Nb_Xpfloat sum_x=0.,sum_y=0.,sum_z=0.;
  const UINT4m size=Size();
  for(UINT4m i=0;i<size;i++) {
    REAL8m mult = Volume(i) * scale[i];
    sum_x += mult*vec[i].x;
    sum_y += mult*vec[i].y;
    sum_z += mult*vec[i].z;
  }

  return ThreeVector(sum_x.GetValue(),sum_y.GetValue(),sum_z.GetValue());
}

