/* FILE: output.h                 -*-Mode: c++-*-
 *
 * Output classes.  Each class providing output should embed an
 * instance of an output class for each output provided.
 *
 * NOTE: This code looks pretty fragile, because there is a lot
 *  of coupling between the output classes and their owner class.
 *  Maybe we can figure out a cleaner, more robust solution in
 *  the future.
 */

#ifndef _OXS_OUTPUT
#define _OXS_OUTPUT

#include "oc.h"
#include "simstate.h"
#include "oxsexcept.h"

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported

/* End includes */

class Oxs_Director;   // Forward declarations

class Oxs_Output {
private:
  Oxs_Director* director;  // Director instance this Oxs_Output
  /// instance is registered with.

  string owner_name;
  string output_name;
  string output_type;
  string output_units;

  BOOL cache_support; // If true, then caching is supported
  UINT4m cache_request_count;

  const INT4m priority; // Output evaluation priority.  Smaller
  /// is higher priority.  Default priority is 0.

  // Disable copy constructor by declaring it without defining it.
  Oxs_Output(const Oxs_Output&);

protected:
  BOOL ForceCacheIncrement(INT4m incr);
  /// Analogous to CacheRequestIncrement, but ignores
  /// cache_support value.  Returns 1 if increment was
  /// successful.  For use by child classes.

  string GetDefaultOutputFormatString() const;
  const char* GetProblemName() const;

  BOOL Setup(const char* owner_name_,
	     const char* output_name_,
	     const char* output_type_,
	     const char* output_units_,
	     BOOL cache_support_,
	     Oxs_Director* director);
  // Registers with director_.  Automatic deregistration is set up
  // in ~Oxs_Output.  Also resets cache_request_count to 0.

  Oxs_Director* GetDirector() const { return director; }

public:
  Oxs_Output(int priority_);
  virtual ~Oxs_Output(); // Deregisters

  string OwnerName()  const { return owner_name;  }
  string OutputName() const { return output_name; }
  string LongName() const;
  string OutputType() const { return output_type; }
  string OutputUnits() const { return output_units; }
  virtual int SetOutputFormat(const string& format) =0;
  virtual string GetOutputFormat() const =0;

  virtual int Output(const Oxs_SimState*,Tcl_Interp*,int argc,CONST84 char** argv)
    =0;

  // Cache flag control.  Returns 1 if increment was successful.
  // NB: The controlling cache_request_count variable gets reset
  //  each time Setup is called.  So CacheRequestIncrement should
  //  be called only _after_ Setup, and cache request needs must
  //  be re-requested any time Setup is re-called.
  BOOL CacheRequestIncrement(INT4m incr)
  {
    if(!cache_support) return 0; // Caching disabled
    return ForceCacheIncrement(incr);
  }

  UINT4m GetCacheRequestCount() const { return cache_request_count; }

  BOOL IsCacheSupported() const { return cache_support; }

  INT4m Priority() const { return priority; }
};

//////////////////////////////////////////////////////////////////
// Templated cache class, used by templated output classes.
template<class T> struct Oxs_OutputCache {
public:
  Oxs_OutputCache() : state_id(0) {}
  UINT4m state_id;
  T value;
};

//////////////////////////////////////////////////////////////////
// Templated output classes.  The templated class T should be the
// embedding class.

////////////////////////
// Scalar output
template<class T> class Oxs_ScalarOutput:public Oxs_Output {
  typedef void (T::* GetScalar)(const Oxs_SimState&);
private:
  T* owner;
  GetScalar getscalar;
  /// Without the typedef, this would be
  ///    void (T::* getscalar)(const Oxs_SimState&)

  string output_format; // Default is "%.17g"

  // Disable copy constructor by declaring it without defining it.
  Oxs_ScalarOutput(const Oxs_ScalarOutput<T> &);

public:
  Oxs_ScalarOutput(INT4m priority_=0) : Oxs_Output(priority_),
    owner(NULL), getscalar(NULL), output_format("%.17g")
    {
      cache.value=0; // Safety
      CacheRequestIncrement(1);  // Temporary measure: until /**/
      /// cache management is implemented in Tcl interface,  /**/
      /// enable caching on all scalar outputs.              /**/
    }

  BOOL Setup(T* owner_,
	     const char* owner_name_,const char* output_name_,
	     const char* output_units_,
	     BOOL cache_support_,const GetScalar member_function,
	     Oxs_Director* director_)
    {
      owner = owner_;
      getscalar = member_function;
      BOOL result
	= Oxs_Output::Setup(owner_name_,output_name_,
			    "scalar",output_units_,
			    cache_support_,director_);
      if(GetDirector()!=NULL) {
	SetOutputFormat(GetDefaultOutputFormatString());
      }
      return result;
    }

  // Note: Base class member function GetDefaultOutputFormatString()
  // calls the director (if any) to get the output format strings
  // stored in the MIF file driver block.  We can't directly access
  // member functions of director_ here because this header file
  // defines a forward declaration to the Oxs_Director class, but does
  // not #include "director.h".  We don't #include "director.h"
  // because director.h needs to include this file (output.h) in order
  // to get the definitions for the templated Oxs_Output child classes
  // Oxs_ScalarOutput and Oxs_VectorFieldOutput (which it uses to
  // provide total energy, total field and energy calc counts).
  // Moreover, we can't move this constructor definition to output.cc
  // because because of the template instantiation model being used
  // (what gcc calls the "Borland model"), which requires the
  // inclusion of full template definitions in each module using the
  // template.  OTOH, the base class Oxs_Output is *not* templated, so
  // Oxs_Output::GetDefaultOutputFormatString() *can* be located in
  // output.cc, which *can* #include "director.h", and so *can* call
  // director->GetDefaultOutputFormatString().

  // Public data member for result drop-off/storage.
  // The owner should test GetCacheRequestCount() inside the output
  // calculation code, and adjust the cache as needed.
  Oxs_OutputCache<REAL8m> cache;

  virtual int SetOutputFormat(const string& format);
  virtual string GetOutputFormat() const { return output_format; }

  virtual int Output(const Oxs_SimState* state,Tcl_Interp* interp,
		     int argc,CONST84 char** argv);
  // The owner getscalar() function can assume on entry that the
  // cache value stored here is invalid, but it may become valid
  // during the lifetime of the getscalar() call (if getscalar()
  // calls some other function that sets the cache).
};

////////////////////////
// Vector field output
template<class T> class Oxs_VectorFieldOutput:public Oxs_Output {
  typedef void (T::* GetVectorField)(const Oxs_SimState&);
private:
  T* owner;
  GetVectorField getvectorfield;
  /// Without the typedef, this would be
  ///    void (T::* getvectorfield)(const Oxs_SimState&)

  string output_format_datatype;  // Default is "binary"
  string output_format_precision; // Default is "8"

  // Disable copy constructor by declaring it without defining it.
  Oxs_VectorFieldOutput(const Oxs_VectorFieldOutput<T> &);

public:

  Oxs_VectorFieldOutput(INT4m priority_=0) : Oxs_Output(priority_),
    owner(NULL), getvectorfield(NULL),
    output_format_datatype("binary"), output_format_precision("8")
    {}

  BOOL Setup(T* owner_,
	     const char* owner_name_,const char* output_name_,
	     const char* output_units_,
	     BOOL cache_support_,
	     const GetVectorField member_function,
	     Oxs_Director* director_)
    {
      owner = owner_;
      getvectorfield = member_function;
      BOOL result
	= Oxs_Output::Setup(owner_name_,output_name_,
			    "vector field",output_units_,
			    cache_support_,director_);
      if(GetDirector()!=NULL) {
	SetOutputFormat(GetDefaultOutputFormatString());
      }
      return result;
    }

  // Note: Base class member function GetDefaultOutputFormatString()
  // calls the director (if any) to get the output format strings
  // stored in the MIF file driver block.  We can't directly access
  // member functions of director_ here because this header file
  // defines a forward declaration to the Oxs_Director class, but does
  // not #include "director.h".  We don't #include "director.h"
  // because director.h needs to include this file (output.h) in order
  // to get the definitions for the templated Oxs_Output child classes
  // Oxs_ScalarOutput and Oxs_VectorFieldOutput (which it uses to
  // provide total energy, total field and energy calc counts).
  // Moreover, we can't move this constructor definition to output.cc
  // because because of the template instantiation model being used
  // (what gcc calls the "Borland model"), which requires the
  // inclusion of full template definitions in each module using the
  // template.  OTOH, the base class Oxs_Output is *not* templated, so
  // Oxs_Output::GetDefaultOutputFormatString() *can* be located in
  // output.cc, which *can* #include "director.h", and so *can* call
  // director->GetDefaultOutputFormatString().

  // Public data member for result drop-off/storage.
  // The owner should test GetCacheRequestCount() inside the output
  // calculation code, and adjust the cache as needed.
  Oxs_OutputCache<Oxs_MeshValue<ThreeVector> > cache;

  virtual int SetOutputFormat(const string& format);
  virtual string GetOutputFormat() const;

  virtual int Output(const Oxs_SimState* state,Tcl_Interp* interp,
		     int argc,CONST84 char** argv);
};

////////////////////////////////////////////////////////////////////////
// Include definitions below.  This is required by compiler/linkers
// using what gcc calls the "Borland model," as opposed to the "Cfront
// model."  The latter is distinguished by the use of template
// repositories.  With the Cfront model, template definitions can be
// put in a file (*.cc) separate from the header file (*.h).  Putting
// the definitions in the header file along with declarations slows down
// compilation, because the definitions are re-compiled in each including
// file, but appears to be more portable.

//////////////////////////////////////////////////////////////
// Oxs_ScalarOutput<T> definitions
template<class T>
int Oxs_ScalarOutput<T>::SetOutputFormat
(const string& format)
{
  if(format.length()==0) return 1;
  output_format = format;
  return 0;
}

template<class T>
int Oxs_ScalarOutput<T>::Output
(const Oxs_SimState* state,   // State to get output for
 Tcl_Interp* interp,          // Tcl interpreter
 int argc,CONST84 char** /* argv */)  // Args
{
  Tcl_ResetResult(interp);
  if(argc!=0) {
    Tcl_AppendResult(interp, "wrong # of args: should be none",
		     (char *) NULL);
    return TCL_ERROR;
  }
  if(state==NULL || state->Id()==0) {
    Tcl_AppendResult(interp,"Invalid state passed to output object",
		     (char *)NULL);
    return TCL_ERROR;
  }

  if(!IsCacheSupported() || cache.state_id!=state->Id()) {
    // Fill cache with data
    ForceCacheIncrement(1);
    cache.state_id=0; // Safety
    (owner->*getscalar)(*state);
    ForceCacheIncrement(-1);

    // Check that cache got filled
    if(cache.state_id!=state->Id()) {
      OXS_THROW(Oxs_ProgramLogicError,
	       "Output cache not filled by registered getscalar"
	       " call (as required).");
    }
  }

  char buf[100];
  Oc_Snprintf(buf,sizeof(buf),
	      output_format.c_str(),cache.value);

  Tcl_ResetResult(interp); // Protection against badly behaved Oxs_Ext's.
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}


//////////////////////////////////////////////////////////////
// Oxs_VectorFieldOutput<T> definitions
template<class T>
int Oxs_VectorFieldOutput<T>::SetOutputFormat
(const string& format)
{
  if(format.length()==0) return 1;
  Oxs_SplitList arglist;
  arglist.Split(format.c_str());
  if(arglist.Count()!=2) {
    OXS_THROW(Oxs_BadParameter,
	      "Wrong number of arguments in format string import to "
	      "Oxs_VectorFieldOutput<T>::SetOutputFormat(const char*)");
  }
  output_format_datatype = arglist[0];
  output_format_precision = arglist[1];
  return 0;
}

template<class T>
string Oxs_VectorFieldOutput<T>::GetOutputFormat() const
{
  vector<string> sarr;
  sarr.push_back(output_format_datatype);
  sarr.push_back(output_format_precision);
  return Oxs_MergeList(&sarr);
}

template<class T>
int Oxs_VectorFieldOutput<T>::Output
(const Oxs_SimState* state, // State to get output for
 Tcl_Interp* interp,        // Tcl interpreter
 int argc,CONST84 char** argv)      // Args
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Tcl_AppendResult(interp, "wrong # of args: should be 1: filename",
		     (char *) NULL);
    return TCL_ERROR;
  }
  if(state==NULL || state->Id()==0) {
    Tcl_AppendResult(interp,"Invalid state passed to output object",
		     (char *)NULL);
    return TCL_ERROR;
  }

  const char* filename = argv[0];

  if(!IsCacheSupported() || cache.state_id!=state->Id()) {
    // Fill cache with vector field
    ForceCacheIncrement(1);
    cache.state_id=0; // Safety
    cache.value.AdjustSize(state->mesh);
    (owner->*getvectorfield)(*state);
    ForceCacheIncrement(-1);

    // Check that cache got filled
    if(cache.state_id!=state->Id()) {
      OXS_THROW(Oxs_ProgramLogicError,
	       "Output cache not filled by registered getvectorfield"
	       " call (as required).");
    }
  }

  // Write cache to file
  char description[4096];
  Oc_Snprintf(description,sizeof(description),
	      "Oxs %.80s output\n"
	      " MIF source file: %.256s\n"
	      " Iteration: %u, State id: %u\n"
	      " Stage: %u, Stage iteration: %u\n"
	      " Stage simulation time: %g s\n"
	      " Total simulation time: %g s\n",
	      OutputType().c_str(),
	      GetProblemName(),
	      state->iteration_count,state->Id(),
	      state->stage_number,state->stage_iteration_count,
	      state->stage_elapsed_time,
	      state->stage_start_time + state->stage_elapsed_time);
  int errorcode = TCL_OK;
  try {
    state->mesh->WriteOvf(filename,
			  LongName().c_str(),
			  description,
			  OutputUnits().c_str(),
			  "rectangular",
			  output_format_datatype.c_str(),
			  output_format_precision.c_str(),
			  &cache.value,NULL);
  } catch (Oxs_Ext::Error& err) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Error writing Ovf file %s, state id %u: ",
		filename,state->Id());
    Tcl_ResetResult(interp); // Protection against badly behaved Oxs_Ext's.
    Tcl_AppendResult(interp,buf,err.c_str(),(char *)NULL);
    errorcode = TCL_ERROR;
  } catch (...) {
    // Unexpected exception; Clean-up and rethrow.
    if(GetCacheRequestCount()<1) {
      cache.state_id=0;
      cache.value.Release();
    }
    throw;
  }

  // Release buffer if caching is disabled
  if(GetCacheRequestCount()<1) {
    cache.state_id=0;
    cache.value.Release();
  }

  // On success, return value is the filename.
  if(errorcode==TCL_OK) {
    Tcl_ResetResult(interp); // Protection against badly behaved Oxs_Ext's.
    Tcl_AppendResult(interp,filename,(char *)NULL);
  }
  return errorcode;
}


#endif // _OXS_OUTPUT


