/* FILE: tclcommand.h                 -*-Mode: c++-*-
 *
 * Wrapper for Tcl_Eval().
 *
 */

#ifndef _OXS_TCLCOMMAND
#define _OXS_TCLCOMMAND

#include <tcl.h>
#include <string>
#include <vector>

#include "oc.h"
#include "util.h"

// Oxs_TclCommand is a wrapper for Tcl_Eval.  The script to be evaluated
// is divided into two parts: the base command and the args.  Setting
// the args and evaluating the command are viewed as a function call on
// the base command, and as such are conceptually const.  For this
// reason, the args and result are stored in mutable storage, so that
// they may be changed by a const Oxs_TclCommand object.
//   For similar reasons, saving and restoring the interp result is
// also considered conceptually const.
class Oxs_TclCommand
{
private:
  string exception_prefix;
  Tcl_Interp* interp;
  mutable vector<Tcl_SavedResult> result_stack;
  mutable BOOL no_too_few_restores_warning; // If this is set,
  /// then no warning is raised on cleanup if there are unpopped
  /// results.  This is intended for exception handling.  It
  /// resets to 0 after each cleanup.
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  string command_base;  // Non-mutable!
  mutable vector<string> command_args;
  mutable string eval_result;
  mutable BOOL eval_list_set;
  mutable Oxs_SplitList eval_list; // Cached list version of eval_result
  void FillEvalList() const;
#else
  int base_command_size;
  Oxs_TclObjArray* objcmdptr;
  mutable Tcl_Obj* eval_result;
#endif // Tcl version check
  int extra_args_count;
public:
  Oxs_TclCommand();
  ~Oxs_TclCommand();

  void SetBaseCommand(const char* exception_prefix_,
		      Tcl_Interp* interp_,
		      const string& cmdbase,
		      int extra_args_count_);
  void ReleaseBaseCommand();

  void SaveInterpResult() const;
  void RestoreInterpResult() const;
  void DiscardInterpResult() const;

  void SetCommandArg(int index,const char* arg) const;
  void SetCommandArg(int index,const string& arg) const
    { SetCommandArg(index,arg.c_str()); }
  void SetCommandArg(int index,INT4m arg) const;
  void SetCommandArg(int index,UINT4m arg) const;
  void SetCommandArg(int index,REAL8m arg) const;

  void Eval() const;

  string GetWholeResult() const;

  int GetResultListSize() const;
  void GetResultListItem(int index,string& result) const;
  void GetResultListItem(int index,INT4m& result) const;
  void GetResultListItem(int index,UINT4m& result) const;
  void GetResultListItem(int index,REAL8m& result) const;
  void GetResultList(vector<string>& result_list) const;
};


/////////////////////////////////////////////////////////////////////
//
// Oxs_ParseTclCommandLineRequest, intended to be used in support of
// Oxs_TclCommand. The label and param_count fields of each
// Oxs_TclCommandLineOptions object in the import/export "options"
// parameter should be filled on entry. Import request_string is
// a Tcl list, where each list item matches the label portion of
// some Oxs_TclCommandLineOptions object in "options".
// Oxs_ParseTclCommandLineRequest fills in the position field of
// each Oxs_TclCommandLineOptions object corresponding to the
// position of that option in the request_string.  If the option
// is not used, the position field is set to -1.  The position
// is padded to account for the number of parameters in each
// preceding option in request_string.  The return value is the
// number of total parameters associated with the options selected
// in request_string.  Exceptions are thrown if request_string
// contains an option not included in "options", or if the same
// options is selected more than once.
//
class Oxs_TclCommandLineOption
{
private:
  string label; // Identifying string for option
  int param_count; // Number of parameters associated with option
  /// I'd like to make label and param_count const, but that would
  /// disallow operator=, which is needed if the class is used
  /// inside STL containers such as vector.
public:
  const string& Label() { return label; }
  int ParamCount() { return param_count; }
  int position; // Command line position; -1 indicates option
  /// not used.
  Oxs_TclCommandLineOption() {};
  Oxs_TclCommandLineOption(string label_,int param_count_)
    : label(label_), param_count(param_count_), position(-1) {}
};
BOOL operator<(const Oxs_TclCommandLineOption&, const Oxs_TclCommandLineOption&);
BOOL operator>(const Oxs_TclCommandLineOption&, const Oxs_TclCommandLineOption&);
BOOL operator==(const Oxs_TclCommandLineOption&, const Oxs_TclCommandLineOption&);
BOOL operator<(const Tcl_SavedResult&, const Tcl_SavedResult&);
BOOL operator>(const Tcl_SavedResult&, const Tcl_SavedResult&);
BOOL operator==(const Tcl_SavedResult&, const Tcl_SavedResult&);

int Oxs_ParseTclCommandLineRequest
(const char* exception_prefix,
 vector<Oxs_TclCommandLineOption>& options,
 const string& request_string);

#endif // _OXS_TCLCOMMAND



