/* FILE: atlasvectorfield.cc      -*-Mode: c++-*-
 *
 * Atlas vector field object, derived from Oxs_VectorField
 * class.
 *
 */

#include <string>
#include <vector>

#include "oc.h"
#include "nb.h"

#include "util.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "atlas.h"
#include "atlasvectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_AtlasVectorField);

/* End includes */

// Constructor
Oxs_AtlasVectorField::Oxs_AtlasVectorField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorField(name,newdtr,argstr)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas_obj);
  atlas_key.Set(atlas_obj.GetPtr());
  atlas_key.GetReadReference(); // Read lock held until *this is deleted.

  BOOL set_norm=0;
  REAL8m norm=0.;
  if(HasInitValue("norm")) {
    norm = GetRealInitValue("norm");
    set_norm = 1;
  }

  // Optional "default_value"; default is (0.,0.,0.)
  OxsAVFSpecifiedValue default_value;
  if(HasInitValue("default_value")) {
    default_value.value = GetThreeVectorInitValue("default_value");
    default_value.isset = 1;
    if(set_norm) default_value.value.SetMag(norm);
  }

  // Fill out value vector with default_value.  This is essentially
  // an empty padding if default.isset==0.
  unsigned int i;
  for(i=0;i<atlas_obj->GetRegionCount();i++) {
    values.push_back(default_value);
  }

  // Get and fill value vector with non-default values
  vector<string> params;
  FindRequiredInitValue("values",params);
  if(params.size()%2!=0) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"List associated with parameter \"values\" has"
		" %d entries, which is not an even number.",
		params.size());
    throw Oxs_Ext::Error(this,buf);
  }
  Oxs_SplitList veclist;
  for(i=0;i<params.size();i+=2) {
    const string& name = params[i];
    INT4m region_id = atlas_obj->GetRegionId(name);
    if(region_id<0) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Specified name, \"%s\","
		  " is not a known region in atlas \"%s\"",
		  name.c_str(),atlas_obj->InstanceName());
      throw Oxs_Ext::Error(this,buf);
    }
    veclist.Split(params[i+1].c_str());
    BOOL error1=1,error2=1,error3=1;
    if(veclist.Count()==3) {
      values[region_id].value =
	ThreeVector(Nb_Atof(veclist[0],error1),
		    Nb_Atof(veclist[1],error2),
		    Nb_Atof(veclist[2],error3));
      if(set_norm) values[region_id].value.SetMag(norm);
    }
    if(error1 || error2 || error3) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Value \"%s\" associated with region \"%s\""
		  " is not a triplet of 3 numbers.",
		  name.c_str(),atlas_obj->InstanceName());
      throw Oxs_Ext::Error(this,buf);
    } else {
      values[region_id].isset = 1;
    }
  }
  DeleteInitValue("values");

  VerifyAllInitArgsUsed();
}

BOOL Oxs_AtlasVectorField::Init()
{ return 1; }

void
Oxs_AtlasVectorField::Value
(const ThreeVector& pt,
 ThreeVector& export_value) const
{
  const Oxs_Atlas* atlas = atlas_key.GetPtr();
  INT4m id = atlas->GetRegionId(pt);
  if(!values[id].isset) {
    char buf[1024];
    string regionname;
    atlas->GetRegionName(id,regionname);
    Oc_Snprintf(buf,sizeof(buf),
		"Error in function Value():"
		" input point (%g,%g,%g), in region %s of atlas %s,"
		" has not been assigned a value.",
		pt.x,pt.y,pt.z,
		regionname.c_str(),atlas->InstanceName());
    throw Oxs_Ext::Error(this,buf);
  }
  export_value = values[id].value;
}
