/* FILE: filevectorfield.cc      -*-Mode: c++-*-
 *
 * File vector field object, derived from Oxs_VectorField class.
 *
 */

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "atlas.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "filevectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_FileVectorField);

/* End includes */


// Constructor
Oxs_FileVectorField::Oxs_FileVectorField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorField(name,newdtr,argstr),
    file_mesh(NULL),set_norm(0),norm(0.0)
{
  // Process arguments.
  if(HasInitValue("norm")) {
    norm = GetRealInitValue("norm");
    set_norm = 1;
  }
  multiplier = GetRealInitValue("multiplier",1.0);
  string filename=GetStringInitValue("file");

  // Determine bounding box.
  Oxs_Box bbox;
  if(HasInitValue("atlas")) {
    // Get bounding box from specified atlas.
    Oxs_OwnedPointer<Oxs_Atlas> atlas;
    OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas);
    atlas->GetWorldExtents(bbox);
  } else {
    // Atlas not specified, so check for and use x/y/zrange data.
    if(!HasInitValue("xrange") || !HasInitValue("yrange")
       || !HasInitValue("zrange")) {
      throw Oxs_Ext::Error(this,
			   "Neither atlas nor x/y/zrange data specified.");
    }
    vector<REAL8m> xrange,yrange,zrange;
    GetRealVectorInitValue("xrange",2,xrange);
    GetRealVectorInitValue("yrange",2,yrange);
    GetRealVectorInitValue("zrange",2,zrange);
    if(!bbox.CheckOrderSet(xrange[0],xrange[1],
			   yrange[0],yrange[1],
			   zrange[0],zrange[1])) {
      throw Oxs_Ext::Error(this,"Bad order in x/y/zrange data.");
    }
  }
  VerifyAllInitArgsUsed();

  // Create and load Vf_Mesh object with file data
  Vf_FileInput* vffi=Vf_FileInput::NewReader(filename.c_str());
  if(vffi==NULL) {
    string msg=string("Error detected during construction of ")
      + string(InstanceName()) + string("; Error reading input file ")
      + filename;
    throw Oxs_Ext::Error(msg.c_str());
  }
  file_mesh=vffi->NewMesh();
  delete vffi;
  if(file_mesh==NULL) {
    string msg=string("Error detected during construction of ")
      + string(InstanceName()) + string(" on input file ")
      + filename + string("; Unable to create Vf_Mesh object.");
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Determine scaling
  REAL8m xmin = bbox.GetMinX();
  REAL8m ymin = bbox.GetMinY();
  REAL8m zmin = bbox.GetMinZ();
  REAL8m xrange = bbox.GetMaxX() - xmin;
  REAL8m yrange = bbox.GetMaxY() - ymin;
  REAL8m zrange = bbox.GetMaxZ() - zmin;

  Nb_BoundingBox<REAL8> file_box;
  file_mesh->GetPreciseRange(file_box);
  Nb_Vec3<REAL8> file_basept;
  file_box.GetMinPt(file_basept);
  REAL8m file_xmin = file_basept.x;
  REAL8m file_ymin = file_basept.y;
  REAL8m file_zmin = file_basept.z;
  REAL8m file_xrange = file_box.GetWidth();
  REAL8m file_yrange = file_box.GetHeight();
  REAL8m file_zrange = file_box.GetDepth();

  if(xrange>0.) {
    scale.x = file_xrange/xrange;
    offset.x = file_xmin - xmin*scale.x;
  } else {
    scale.x = 0.0;
    offset.x = file_xmin/2.0;
  }
  if(yrange>0.) {
    scale.y = file_yrange/yrange;
    offset.y = file_ymin - ymin*scale.y;
  } else {
    scale.y = 0.0;
    offset.y = file_ymin/2.0;
  }
  if(zrange>0.) {
    scale.z = file_zrange/zrange;
    offset.z = file_zmin - zmin*scale.z;
  } else {
    scale.z = 0.0;
    offset.z = file_zmin/2.0;
  }

  valuescale = file_mesh->GetValueMultiplier();
  valuescale *= multiplier; // For efficiency, roll multiplier
  /// into valuescale.  Note that valuescale is only used if set_norm
  /// is false.  In the case where set_norm is true, multiplier
  /// is explicitly applied after renormalization.
}

Oxs_FileVectorField::~Oxs_FileVectorField()
{
  if(file_mesh!=NULL) delete file_mesh;
}


BOOL Oxs_FileVectorField::Init()
{ return 1; }

void
Oxs_FileVectorField::Value
(const ThreeVector& pt,
 ThreeVector& value) const
{
  Nb_Vec3<REAL8> workpt(offset.x+pt.x*scale.x,
			offset.y+pt.y*scale.y,
			offset.z+pt.z*scale.z);
  Nb_LocatedVector<REAL8> lv;
  file_mesh->FindPreciseClosest(workpt,lv);

  if(set_norm) {
    value.x=lv.value.x;
    value.y=lv.value.y;
    value.z=lv.value.z;
    if(norm==1.0) value.MakeUnit();
    else          value.SetMag(norm);
    if(multiplier!=1.0) value *= multiplier;
  } else {
    // Note: For efficiency, multiplier is already rolled into
    // valuescale.
    value.x=valuescale*lv.value.x;
    value.y=valuescale*lv.value.y;
    value.z=valuescale*lv.value.z;
  }

}

void
Oxs_FileVectorField::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<ThreeVector>& array) const
{
  array.AdjustSize(mesh);

  // Variables for file_mesh (Vf_Mesh*) interface
  Nb_Vec3<REAL8> file_pos;  Nb_LocatedVector<REAL8> file_lv;
  file_lv.value.Set(1.,0.,0.); // Safety

  // Variables for mesh (Oxs_Mesh*) interface
  ThreeVector center,value;

  // Step through range and sample mesh, fitting
  // with zeroth order fit.
  UINT4m size=mesh->Size();
  for(UINT4m i=0;i<size;i++) {
    mesh->Center(i,center);
    center.x *= scale.x; center.y *= scale.y; center.z *= scale.z;
    center += offset;
    file_pos.x=center.x;
    file_pos.y=center.y;
    file_pos.z=center.z;
    file_mesh->FindPreciseClosest(file_pos,file_lv);

    if(set_norm) {
      value.x=file_lv.value.x;
      value.y=file_lv.value.y;
      value.z=file_lv.value.z;
      if(norm==1.0) value.MakeUnit();
      else          value.SetMag(norm);
      if(multiplier!=1.0) value *= multiplier;
    } else {
      // Note: For efficiency, multiplier is already rolled into
      // valuescale.
      value.x=valuescale*file_lv.value.x;
      value.y=valuescale*file_lv.value.y;
      value.z=valuescale*file_lv.value.z;
    }
    array[i]=value;
  }

}


